# Licensed under a 3-clause BSD style license - see LICENSE.rst
from .. import beam as radio_beam
from astropy import units as u
import numpy.testing as npt
import numpy as np
import pytest
from pkg_resources import parse_version
from astropy.version import version

SIGMA_TO_FWHM = radio_beam.SIGMA_TO_FWHM

min_astropy_version = parse_version("1.1")


@pytest.mark.skipif(parse_version(version) < min_astropy_version,
                    reason="Must have astropy version >1.1")
def test_gauss_kernel():

    fake_beam = radio_beam.Beam(10)

    # Let pixscale be 0.1 deg/pix
    kernel = fake_beam.as_kernel(0.1*u.deg)

    direct_kernel = \
        radio_beam.EllipticalGaussian2DKernel(100. / SIGMA_TO_FWHM,
                                              100. / SIGMA_TO_FWHM,
                                              0.0)

    npt.assert_allclose(kernel.array, direct_kernel.array)


@pytest.mark.skipif(parse_version(version) < min_astropy_version,
                    reason="Must have astropy version >1.1")
def test_tophat_kernel():

    fake_beam = radio_beam.Beam(10)

    # Let pixscale be 0.1 deg/pix
    kernel = fake_beam.as_tophat_kernel(0.1*u.deg)

    direct_kernel = \
        radio_beam.EllipticalTophat2DKernel(100. / (SIGMA_TO_FWHM/np.sqrt(2)),
                                            100. / (SIGMA_TO_FWHM/np.sqrt(2)),
                                            0.0)

    npt.assert_allclose(kernel.array, direct_kernel.array)




import math
import numpy as np
import astropy.units as u


DEG2RAD = math.pi / 180.

class BeamError(Exception):
    """docstring for BeamError"""
    pass


class InvalidBeamOperationError(Exception):
    pass


class RadioBeamDeprecationWarning(Warning):
    pass


def deconvolve_optimized(beamprops1, beamprops2, failure_returns_pointlike=False):
    """
    An optimized, non-Quantity version of beam deconvolution.

    Because no unit conversions are handled, the inputs MUST be in degrees for the major, minor,
    and position angle.

    Parameters
    ----------
    beamprops1: dict
        Dictionary with keys 'BMAJ', 'BMIN', and 'BPA' for the beam to deconvolve from. Can be
        produced with `~radio_beam.Beam.to_fits_keywords`.
    beamprops2: dict
        Same as `beamprops1` for the second beam.
    failure_returns_pointlike : bool, optional
        Return a point beam (zero area) when deconvolution fails. If `False`,
        this will instead raise a `~radio_beam.utils.BeamError` when deconvolution fails.


    Returns
    -------
    new_major : float
        Deconvolved major FWHM.
    new_minor : float
        Deconvolved minor FWHM.
    new_pa : float
        Deconvolved position angle.

    """

    # blame: https://github.com/pkgw/carma-miriad/blob/CVSHEAD/src/subs/gaupar.for
    # (githup checkin of MIRIAD, code by Sault)

    maj1 = beamprops1['BMAJ']
    min1 = beamprops1['BMIN']
    pa1 = beamprops1['BPA'] * DEG2RAD

    maj2 = beamprops2['BMAJ']
    min2 = beamprops2['BMIN']
    pa2 = beamprops2['BPA'] * DEG2RAD

    alpha = ((maj1 * math.cos(pa1))**2 +
             (min1 * math.sin(pa1))**2 -
             (maj2 * math.cos(pa2))**2 -
             (min2 * math.sin(pa2))**2)

    beta = ((maj1 * math.sin(pa1))**2 +
            (min1 * math.cos(pa1))**2 -
            (maj2 * math.sin(pa2))**2 -
            (min2 * math.cos(pa2))**2)

    gamma = 2 * ((min1**2 - maj1**2) * math.sin(pa1) * math.cos(pa1) -
                 (min2**2 - maj2**2) * math.sin(pa2) * math.cos(pa2))

    s = alpha + beta
    t = math.sqrt((alpha - beta)**2 + gamma**2)

    # Deal with floating point issues
    # This matches the arcsec**2 check for deconvolve below
    # Difference is we keep things in deg^2 here
    atol_t = np.finfo(np.float64).eps / 3600.**2

    # To deconvolve, the beam must satisfy:
    # alpha < 0
    alpha_cond = alpha + np.finfo(np.float64).eps < 0

    # beta < 0
    beta_cond = beta + np.finfo(np.float64).eps < 0
    # s < t
    st_cond = s < t + atol_t

    if alpha_cond or beta_cond or st_cond:
        if failure_returns_pointlike:
            return 0., 0., 0.
        else:
            raise BeamError("Beam could not be deconvolved")
    else:
        new_major = math.sqrt(0.5 * (s + t))
        new_minor = math.sqrt(0.5 * (s - t))

        # absolute tolerance needs to be <<1 microarcsec
        atol = 1e-7 / 3600.
        if (math.sqrt(abs(gamma) + abs(alpha - beta))) < atol:
            new_pa = 0.0
        else:
            new_pa = 0.5 * math.atan2(-1. * gamma, alpha - beta)

    # In the limiting case, the axes can be zero to within precision
    # Add the precision level onto each axis so a deconvolvable beam
    # is always has beam.isfinite == True
    new_major += np.finfo(np.float64).eps
    new_minor += np.finfo(np.float64).eps

    return new_major, new_minor, new_pa


def deconvolve(beam, other, failure_returns_pointlike=False):
    """
    Deconvolve a beam from another

    Parameters
    ----------
    beam : `Beam`
        The defined beam.
    other : `Beam`
        The beam to deconvolve from this beam
    failure_returns_pointlike : bool
        Option to return a pointlike beam (i.e., one with major=minor=0) if
        the second beam is larger than the first.  Otherwise, a ValueError
        will be raised

    Returns
    -------
    new_beam : `Beam`
        The convolved Beam

    Raises
    ------
    failure : ValueError
        If the second beam is larger than the first, the default behavior
        is to raise an exception.  This can be overridden with
        failure_returns_pointlike
    """


    # The header keywords handle the conversions to degree for BMAJ, BMIN, BPA.
    beamprops1 = beam.to_header_keywords()
    beamprops2 = other.to_header_keywords()

    return deconvolve_optimized(beamprops1, beamprops2,
                                failure_returns_pointlike=failure_returns_pointlike)


def convolve(beam, other):
    """
    Convolve one beam with another.

    Parameters
    ----------
    other : `Beam`
        The beam to convolve with

    Returns
    -------
    new_beam : `Beam`
        The convolved Beam
    """

    # blame: https://github.com/pkgw/carma-miriad/blob/CVSHEAD/src/subs/gaupar.for
    # (github checkin of MIRIAD, code by Sault)

    alpha = ((beam.major * np.cos(beam.pa))**2 +
             (beam.minor * np.sin(beam.pa))**2 +
             (other.major * np.cos(other.pa))**2 +
             (other.minor * np.sin(other.pa))**2)

    beta = ((beam.major * np.sin(beam.pa))**2 +
            (beam.minor * np.cos(beam.pa))**2 +
            (other.major * np.sin(other.pa))**2 +
            (other.minor * np.cos(other.pa))**2)

    gamma = (2 * ((beam.minor**2 - beam.major**2) *
                  np.sin(beam.pa) * np.cos(beam.pa) +
                  (other.minor**2 - other.major**2) *
                  np.sin(other.pa) * np.cos(other.pa)))

    s = alpha + beta
    t = np.sqrt((alpha - beta)**2 + gamma**2)

    new_major = np.sqrt(0.5 * (s + t))
    new_minor = np.sqrt(0.5 * (s - t))
    # absolute tolerance needs to be <<1 microarcsec
    if np.isclose(((abs(gamma) + abs(alpha - beta))**0.5).to(u.arcsec).value, 1e-7):
        new_pa = 0.0 * u.deg
    else:
        new_pa = 0.5 * np.arctan2(-1. * gamma, alpha - beta)

    return new_major, new_minor, new_pa


def transform_ellipse(major, minor, pa, x_scale, y_scale):
    '''
    Transform an ellipse by scaling in the x and y axes.

    Parameters
    ----------
    major : `~astropy.units.Quantity`
        Major axis.
    minor : `~astropy.units.Quantity`
        Minor axis.
    pa : `~astropy.units.Quantity`
        PA of the major axis.
    x_scale : float
        x axis scaling factor.
    y_scale : float
        y axis scaling factor.

    Returns
    -------
    trans_major : `~astropy.units.Quantity`
        Major axis in the transformed frame.
    trans_minor : `~astropy.units.Quantity`
        Minor axis in the transformed frame.
    trans_pa : `~astropy.units.Quantity`
        PA of the major axis in the transformed frame.
    '''

    # This code is based on the implementation in CASA:
    # https://open-bitbucket.nrao.edu/projects/CASA/repos/casa/browse/code/imageanalysis/ImageAnalysis/CasaImageBeamSet.cc

    major = major.to(u.arcsec)
    minor = minor.to(u.arcsec)
    pa = pa.to(u.rad)

    cospa = np.cos(pa)
    sinpa = np.sin(pa)
    cos2pa = cospa**2
    sin2pa = sinpa**2

    major2 = major**2
    minor2 = minor**2

    a = (cos2pa / major2) + (sin2pa / minor2)
    b = -2 * cospa * sinpa * (major2**-1 - minor2**-1)
    c = (sin2pa / major2) + (cos2pa / minor2)

    x2_scale = x_scale**2
    y2_scale = y_scale**2

    r = a / x2_scale
    s = b**2 / (4 * x2_scale * y2_scale)
    t = c / y2_scale

    udiff = r - t
    u2 = udiff**2

    f1 = u2 + 4 * s
    f2 = np.sqrt(f1) * np.abs(udiff)

    j1 = (f2 + f1) / f1 / 2
    j2 = (f1 - f2) / f1 / 2

    k1 = (j1 * r + j1 * t - t) / (2 * j1 - 1)
    k2 = (j2 * r + j2 * t - t) / (2 * j2 - 1)

    c1 = np.sqrt(k1)**-1
    c2 = np.sqrt(k2)**-1

    pa_sign = 1 if pa.value >= 0 else -1

    if c1 == c2:
        # Transformed to a circle
        trans_major = 1 / c1
        trans_minor = trans_major
        trans_pa = 0. * u.rad

    elif c1 > c2:
        # c1 and c2 are the major and minor axes; use j1 to get PA
        trans_major = c1
        trans_minor = c2
        trans_pa = pa_sign * np.arccos(np.sqrt(j1))
    else:
        # Opposite case where the axes are switched; get PA from j2
        trans_major = c2
        trans_minor = c1
        trans_pa = pa_sign * np.arccos(np.sqrt(j2))

    return trans_major, trans_minor, trans_pa



def get_package_data():
    return {
        _ASTROPY_PACKAGE_NAME_ + '.tests': ['data/*.fits.gz',
                                            'data/*.tar.gz',
                                            'data/*hdr',
                                            'data/*.csv']}



import numpy as np
import numpy.testing as npt

from astropy import units as u
from astropy.io import fits

import warnings
import pytest

from ..multiple_beams import Beams
from ..beam import Beam
from ..commonbeam import common_2beams, common_manybeams_mve
from ..utils import InvalidBeamOperationError, BeamError

from .test_beam import data_path


def symm_beams_for_tests():

    majors = [1, 1, 1, 2, 3, 4] * u.arcsec
    minors = majors
    pas = [0] * 6 * u.deg

    return Beams(major=majors, minor=minors, pa=pas), majors, minors, pas


def asymm_beams_for_tests():

    majors = [1, 1, 1, 2, 3, 4] * u.arcsec
    minors = majors / 2.
    pas = [-36, 20, 80, 41, -82, 11] * u.deg

    return Beams(major=majors, minor=minors, pa=pas), majors, minors, pas


def load_commonbeam_comparisons():

    common_beams = np.loadtxt(data_path("commonbeam_CASA_comparison.csv"),
                              delimiter=',')

    return common_beams


def test_beam_areas():

    beams, majors = symm_beams_for_tests()[:2]

    areas = 2 * np.pi / (8 * np.log(2)) * (majors.to(u.rad)**2).to(u.sr)

    assert np.all(areas.value == beams.sr.value)
    assert np.all(beams.value == beams.sr.value)


def test_beams_from_fits_bintable():

    fname = data_path("m33_beams_bintable.fits.gz")

    bintable = fits.open(fname)[1]

    beams = Beams.from_fits_bintable(bintable)

    assert (beams.major.value == bintable.data['BMAJ']).all()
    assert (beams.minor.value == bintable.data['BMIN']).all()
    assert (beams.pa.value == bintable.data['BPA']).all()


def test_beams_from_list_of_beam():

    beams, majors = symm_beams_for_tests()[:2]

    new_beams = Beams(beams=[beam for beam in beams])

    assert beams == new_beams

    abeams = asymm_beams_for_tests()[0]
    new_abeams = Beams(beams=[beam for beam in abeams])
    assert abeams == new_abeams


def test_beams_equality_beams():

    beams, majors = symm_beams_for_tests()[:2]

    assert beams == beams

    assert not beams != beams

    abeams, amajors = asymm_beams_for_tests()[:2]

    assert not (beams == abeams)

    assert beams != abeams


def test_beams_equality_beam():

    # Test whether all are equal to a single beam
    beams = Beams([1.] * 5 * u.arcsec)

    beam = Beam(1 * u.arcsec)

    assert np.all(beams == beam)

    assert not np.any(beams != beam)


@pytest.mark.xfail(raises=InvalidBeamOperationError, strict=True)
def test_beams_equality_fail():

    # Test whether all are equal to a single beam
    beams = Beams([1.] * 5 * u.arcsec)

    beams == 2


@pytest.mark.xfail(raises=InvalidBeamOperationError, strict=True)
def test_beams_notequality_fail():

    # Test whether all are equal to a single beam
    beams = Beams([1.] * 5 * u.arcsec)

    beams != 2


@pytest.mark.xfail(raises=InvalidBeamOperationError, strict=True)
def test_beams_equality_fail_shape():

    # Test whether all are equal to a single beam
    beams = Beams([1.] * 5 * u.arcsec)

    assert np.all(beams == beams[1:])

@pytest.mark.xfail(raises=InvalidBeamOperationError, strict=True)
def test_beams_add_fail():

    # Test whether all are equal to a single beam
    beams = Beams([1.] * 5 * u.arcsec)

    beams + 2


@pytest.mark.xfail(raises=InvalidBeamOperationError, strict=True)
def test_beams_sub_fail():

    # Test whether all are equal to a single beam
    beams = Beams([1.] * 5 * u.arcsec)

    beams - 2


@pytest.mark.xfail(raises=InvalidBeamOperationError, strict=True)
def test_beams_mult_fail():

    # Test whether all are equal to a single beam
    beams = Beams([1.] * 5 * u.arcsec)

    beams * 2


@pytest.mark.xfail(raises=InvalidBeamOperationError, strict=True)
def test_beams_div_fail():

    # Test whether all are equal to a single beam
    beams = Beams([1.] * 5 * u.arcsec)

    beams / 2


def test_beams_mult_convolution():

    beams, majors = asymm_beams_for_tests()[:2]

    beam = Beam(1 * u.arcsec)

    conv_beams = beams * beam

    individ_conv_beams = [beam_i.convolve(beam) for beam_i in beams]
    new_beams = Beams(beams=individ_conv_beams)

    assert conv_beams == new_beams


def test_beams_div_deconvolution():

    beams, majors = asymm_beams_for_tests()[:2]

    beam = Beam(0.25 * u.arcsec)

    deconv_beams = beams / beam

    individ_deconv_beams = [beam_i.deconvolve(beam) for beam_i in beams]
    new_beams = Beams(beams=individ_deconv_beams)

    assert deconv_beams == new_beams


def test_indexing():

    beams, majors = symm_beams_for_tests()[:2]

    assert hasattr(beams[slice(0, 3)], 'major')
    assert np.all(beams[slice(0, 3)].major.value == majors[:3].value)
    assert np.all(beams[slice(0, 3)].minor.value == majors[:3].value)

    assert hasattr(beams[:3], 'major')
    assert np.all(beams[:3].major.value == majors[:3].value)
    assert np.all(beams[:3].minor.value == majors[:3].value)

    assert hasattr(beams[3], 'major')
    assert beams[3].major.value == 2
    assert beams[3].minor.value == 2
    assert isinstance(beams[4], Beam)

    # Also test int64
    chan = np.int64(3)
    assert hasattr(beams[chan], 'major')
    assert beams[chan].major.value == 2
    assert beams[chan].minor.value == 2
    assert isinstance(beams[chan], Beam)

    mask = np.array([True, False, True, False, True, True], dtype='bool')
    assert hasattr(beams[mask], 'major')
    assert np.all(beams[mask].major.value == majors[mask].value)


def test_average_beams():

    beams, majors = symm_beams_for_tests()[:2]

    assert np.all(beams.average_beam().major.value == majors.mean().value)

    mask = np.array([True, False, True, False, True, True], dtype='bool')

    assert np.all(beams[mask].average_beam().major.value ==
                  majors[mask].mean().value)


@pytest.mark.parametrize(("beams", "majors", "minors", "pas"),
                         [symm_beams_for_tests(), asymm_beams_for_tests()])
def test_largest_beams(beams, majors, minors, pas):

    assert beams.largest_beam().major.value == majors.max().value
    assert beams.largest_beam().minor.value == minors.max().value

    # Slice the object
    mask = np.array([True, False, True, False, True, True], dtype='bool')

    assert beams[mask].largest_beam().major.value == majors[mask].max().value
    assert beams[mask].largest_beam().minor.value == minors[mask].max().value

    # Apply a mask only for the operation
    assert beams.largest_beam(mask).major.value == majors[mask].max().value
    assert beams.largest_beam(mask).minor.value == minors[mask].max().value


@pytest.mark.parametrize(("beams", "majors", "minors", "pas"),
                         [symm_beams_for_tests(), asymm_beams_for_tests()])
def test_smallest_beams(beams, majors, minors, pas):

    assert beams.smallest_beam().major.value == majors.min().value
    assert beams.smallest_beam().minor.value == minors.min().value

    # Slice the object
    mask = np.array([True, False, True, False, True, True], dtype='bool')

    assert beams[mask].smallest_beam().major.value == majors[mask].min().value
    assert beams[mask].smallest_beam().minor.value == minors[mask].min().value

    # Apply a mask only for the operation
    assert beams.smallest_beam(mask).major.value == majors[mask].min().value
    assert beams.smallest_beam(mask).minor.value == minors[mask].min().value


@pytest.mark.parametrize(("beams", "majors", "minors", "pas"),
                         [symm_beams_for_tests(), asymm_beams_for_tests()])
def test_extrema_beams(beams, majors, minors, pas):

    extrema = beams.extrema_beams()
    assert extrema[0].major.value == majors.min().value
    assert extrema[0].minor.value == minors.min().value

    assert extrema[1].major.value == majors.max().value
    assert extrema[1].minor.value == minors.max().value

    # Slice the object
    mask = np.array([True, False, True, False, True, True], dtype='bool')
    extrema = beams[mask].extrema_beams()
    assert extrema[0].major.value == majors[mask].min().value
    assert extrema[0].minor.value == minors[mask].min().value

    assert extrema[1].major.value == majors[mask].max().value
    assert extrema[1].minor.value == minors[mask].max().value

    # Apply a mask only for the operation
    extrema = beams.extrema_beams(mask)
    assert extrema[0].major.value == majors[mask].min().value
    assert extrema[0].minor.value == minors[mask].min().value

    assert extrema[1].major.value == majors[mask].max().value
    assert extrema[1].minor.value == minors[mask].max().value


@pytest.mark.parametrize("majors", [[1, 1, 1, 2, np.NaN, 4],
                                    [0, 1, 1, 2, 3, 4]])
def test_beams_with_invalid(majors):

    majors = np.asarray(majors) * u.arcsec

    beams = Beams(major=majors)

    # Average
    assert beams.average_beam().major.value == np.nanmean(
        majors[np.nonzero(majors)]).value
    # Largest
    assert beams.largest_beam().major.value == np.nanmax(majors).value
    # Smallest
    assert beams.smallest_beam().major.value == np.nanmin(
        majors[np.nonzero(majors)]).value
    # Extrema
    extrema = beams.extrema_beams()
    assert extrema[0].major.value == np.nanmin(
        majors[np.nonzero(majors)]).value
    assert extrema[1].major.value == np.nanmax(majors).value

    # Additional masking
    mask = np.array([True, False, True, False, True, True], dtype='bool')

    if np.isnan(majors).any():
        bad_mask = np.isfinite(majors)
    else:
        bad_mask = majors.value != 0

    combined_mask = np.logical_and(mask, bad_mask)

    # Average
    assert beams[mask].average_beam().major.value == np.nanmean(
        majors[combined_mask]).value
    # Largest
    assert beams[mask].largest_beam().major.value == np.nanmax(
        majors[combined_mask]).value
    # Smallest
    assert beams[mask].smallest_beam().major.value == np.nanmin(
        majors[combined_mask]).value
    # Extrema
    extrema = beams[mask].extrema_beams()
    assert extrema[0].major.value == np.nanmin(majors[combined_mask]).value
    assert extrema[1].major.value == np.nanmax(majors[combined_mask]).value


def test_beams_iter():

    beams, majors = symm_beams_for_tests()[:2]

    # Ensure iterating through yields the same as slicing
    for i, beam in enumerate(beams):
        assert beam == beams[i]


# @pytest.mark.parametrize('comp_vals',
#                          [vals for vals in load_commonbeam_comparisons()])
# def test_commonbeam_casa_compare(comp_vals):

#     # These are the common beam parameters assuming 2 beams:
#     # 1) 3"x3"
#     # 2) 4"x2.5", varying the PA in 1 deg increments from 0 to 179 deg
#     # See data/generate_commonbeam_table.py

#     pa, com_major, com_minor, com_pa = comp_vals

#     beams = Beams(major=[3, 4] * u.arcsec, minor=[3, 2.5] * u.arcsec,
#                   pa=[0, pa] * u.deg)

#     common_beam = beams.common_beam()

#     # npt.assert_almost_equal(common_beam.major.value, com_major)
#     # npt.assert_almost_equal(common_beam.minor.value, com_minor)
#     npt.assert_almost_equal(common_beam.pa.to(u.deg).value, com_pa)
#     npt.assert_almost_equal(common_beam.pa.to(u.deg).value, pa)


def test_common_beam_smallcircular():
    '''
    Simple solution if the smallest beam is circular with a radius larger:
    Major axis is from the largest beam, minor axis is the radius of the
    smaller, and the PA is from the largest beam.
    '''

    for pa in [0., 18., 68., 122.]:
        beams = Beams(major=[3, 4] * u.arcsec,
                      minor=[3, 2.5] * u.arcsec,
                      pa=[0, pa] * u.deg)

        targ_beam = Beam(4 * u.arcsec, 3 * u.arcsec, pa * u.deg)

        assert targ_beam == beams.common_beam()


def test_commonbeam_notlargest():

    beams = Beams(major=[3, 4] * u.arcsec, minor=[3, 2.5] * u.arcsec)

    target_beam = Beam(major=4 * u.arcsec,
                       minor=3 * u.arcsec)

    assert beams.common_beam() == target_beam


def test_commonbeam_largest():
    '''
    commonbeam is the largest in this set.
    '''

    beams, majors = symm_beams_for_tests()[:2]

    assert beams.common_beam() == beams.largest_beam()

    # With masking
    mask = np.array([True, False, True, True, True, False], dtype='bool')

    assert beams[mask].common_beam() == beams[mask].largest_beam()

    assert beams.common_beam(mask) == beams.largest_beam(mask)


# Implements the same test suite used in CASA

def casa_commonbeam_suite():

    cases = []

    # https://open-bitbucket.nrao.edu/projects/CASA/repos/casa/browse/code/imageanalysis/ImageAnalysis/test/tCasaImageBeamSet.cc

    # In some cases, I find smaller common beams than are listed in the CASA
    # tests. The values for the CASA tests are commented out.

    # 1
    cases.append((Beams(major=[4] * 2 * u.arcsec, minor=[2] * 2 * u.arcsec,
                        pa=[0, 60] * u.deg),
                  Beam(major=4.4812 * u.arcsec, minor=3.2883 * u.arcsec,
                       pa=30.0 * u.deg)))
                  # Beam(major=4.4856 * u.arcsec, minor=3.2916 * u.arcsec,
                  #      pa=30.0 * u.deg)))
    # 2
    cases.append((Beams(major=[4] * 2 * u.arcsec, minor=[2] * 2 * u.arcsec,
                        pa=[20, 80] * u.deg),
                  Beam(major=4.4812 * u.arcsec, minor=3.2883 * u.arcsec,
                       pa=50.0 * u.deg)))
                  # Beam(major=4.4856 * u.arcsec, minor=3.2916 * u.arcsec,
                  #      pa=50.0 * u.deg)))
    # 3
    cases.append((Beams(major=[4] * 2 * u.arcsec, minor=[2] * 2 * u.arcsec,
                        pa=[1, 89] * u.deg),
                  Beam(major=4.042 * u.arcsec, minor=3.958 * u.arcsec,
                       pa=45.0 * u.deg)))
    # 4
    cases.append((Beams(major=[4] * 2 * u.arcsec, minor=[2] * 2 * u.arcsec,
                        pa=[0, 90] * u.deg),
                  Beam(major=4 * u.arcsec, minor=4 * u.arcsec,
                       pa=0.0 * u.deg)))
    # 5
    cases.append((Beams(major=[4, 1.5] * u.arcsec, minor=[2, 1] * u.arcsec,
                        pa=[0, 90] * u.deg),
                  Beam(major=4 * u.arcsec, minor=2 * u.arcsec,
                       pa=0.0 * u.deg)))
    # 6
    cases.append((Beams(major=[8, 4] * u.arcsec, minor=[1, 1] * u.arcsec,
                        pa=[0, 20] * u.deg),
                  Beam(major=8.3684 * u.arcsec, minor=1.6253 * u.arcsec,
                       pa=2.7679 * u.deg)))
                  # Beam(major=8.377 * u.arcsec, minor=1.628 * u.arcsec,
                  #      pa=2.7679 * u.deg)))
    # 7
    cases.append((Beams(major=[4, 8] * u.arcsec, minor=[1, 1] * u.arcsec,
                        pa=[0, 20] * u.deg),
                  Beam(major=8.369 * u.arcsec, minor=1.626 * u.arcsec,
                       pa=17.232 * u.deg)))

    # 10
    cases.append((Beams(major=[4, 1] * u.arcsec, minor=[2, 1] * u.arcsec,
                        pa=[0, 0] * u.deg),
                  Beam(major=4 * u.arcsec, minor=2 * u.arcsec,
                       pa=0.0 * u.deg)))

    return cases


@pytest.mark.parametrize(("beams", "target_beam"), casa_commonbeam_suite())
def test_commonbeam_angleoffset(beams, target_beam):

    # https://open-bitbucket.nrao.edu/projects/CASA/repos/casa/browse/code/imageanalysis/ImageAnalysis/test/tCasaImageBeamSet.cc#447

    common_beam = beams.common_beam()

    # Order shouldn't matter
    common_beam_rev = beams[::-1].common_beam()

    assert common_beam == common_beam_rev

    npt.assert_allclose(common_beam.major.value, target_beam.major.value,
                        rtol=1e-3)
    npt.assert_allclose(common_beam.minor.value, target_beam.minor.value,
                        rtol=1e-3)
    npt.assert_allclose(common_beam.pa.to(u.deg).value,
                        target_beam.pa.value, rtol=1e-3)


def casa_commonbeam_suite_multiple():

    cases = []

    # 8
    cases.append((Beams(major=[4] * 4 * u.arcsec, minor=[2] * 4 * u.arcsec,
                        pa=[0, 60, 20, 40] * u.deg),
                  Beam(major=4.48904471492 * u.arcsec,
                       minor=3.28268221138 * u.arcsec,
                       pa=30.0001561178 * u.deg)))

    # This is the beam size in the CASA tests. The MVE method finds a slightly
    # smaller beam area, so that's what is tested against above and below.
                  # Beam(major=4.485 * u.arcsec, minor=3.291 * u.arcsec,
                  #      pa=30 * u.deg)))
    # 9
    cases.append((Beams(major=[4] * 4 * u.arcsec, minor=[2] * 4 * u.arcsec,
                        pa=[0, 20, 40, 60] * u.deg),
                  Beam(major=4.48904471492 * u.arcsec,
                       minor=3.28268221138 * u.arcsec,
                       pa=30.0001561178 * u.deg)))

    return cases


@pytest.mark.parametrize(("beams", "target_beam"),
                         casa_commonbeam_suite_multiple())
def test_commonbeam_multiple(beams, target_beam):

    # https://open-bitbucket.nrao.edu/projects/CASA/repos/casa/browse/code/imageanalysis/ImageAnalysis/test/tCasaImageBeamSet.cc#447

    common_beam = beams.common_beam(epsilon=1e-4)

    # The above should be using the MVE method
    common_beam_check = common_manybeams_mve(beams, epsilon=1e-4)

    assert common_beam == common_beam_check

    npt.assert_almost_equal(common_beam.major.to(u.arcsec).value,
                            target_beam.major.value,
                            decimal=6)
    npt.assert_almost_equal(common_beam.minor.to(u.arcsec).value,
                            target_beam.minor.value,
                            decimal=6)
    npt.assert_allclose(common_beam.pa.to(u.deg).value,
                        target_beam.pa.value, rtol=1e-3)


@pytest.mark.parametrize(("beams", "target_beam"), casa_commonbeam_suite())
def test_commonbeam_methods(beams, target_beam):

    epsilon = 5e-4
    tolerance = 1e-4

    two_beam_method = common_2beams(beams)
    many_beam_method = common_manybeams_mve(beams, epsilon=epsilon,
                                            tolerance=tolerance)

    # Good to ~5x the given epsilon
    npt.assert_allclose(two_beam_method.major.to(u.arcsec).value,
                        many_beam_method.major.to(u.arcsec).value,
                        rtol=3e-3)
    npt.assert_allclose(two_beam_method.minor.to(u.arcsec).value,
                        many_beam_method.minor.to(u.arcsec).value,
                        rtol=3e-3)

    # Only test if the beam is
    circ_check = not two_beam_method.iscircular(rtol=3e-3) or \
        not many_beam_method.iscircular(rtol=3e-3)
    if circ_check:
        # The pa can be sensitive to small changes so give it a larger
        # acceptable tolerance range.
        npt.assert_allclose(two_beam_method.pa.to(u.deg).value,
                            many_beam_method.pa.to(u.deg).value,
                            rtol=5e-3)


def test_catch_common_beam_opt():
    '''
    The optimization method is close to working, but requires more testing.
    Ensure it cannot be used.
    '''

    beams = Beams(major=[4] * 4 * u.arcsec, minor=[2] * 4 * u.arcsec,
                  pa=[0, 20, 40, 60] * u.deg)

    with pytest.raises(NotImplementedError):
        beams.common_beam(method='opt')


def test_major_minor_swap():

    with pytest.raises(ValueError) as exc:
        beams = Beams(minor=[10.,5.] * u.arcsec,
                      major=[5., 5.] * u.arcsec,
                      pa=[30., 60.] * u.deg)

    assert "Minor axis greater than major axis." in exc.value.args[0]


def test_common_beam_mve_auto_increase_epsilon():
    '''
    Here's a case where the default MVE parameters fail.
    By slowly increasing the epsilon* value, we get a common
    beam the can be deconvolved correctly over the set.

    * epsilon is the small factor added to the ellipse perimeter
    radius: radius * (1 + epsilon). The solution is then marginally
    larger than the true optimal solution, but close enough for
    effectively all use cases.
    '''

    major = [8.517199, 8.513563, 8.518497, 8.518434, 8.528561, 8.528236,
             8.530046, 8.530528, 8.530696, 8.533117] * u.arcsec
    minor = [5.7432523, 5.7446027, 5.7407207, 5.740814, 5.7331843,
             5.7356524, 5.7338963, 5.733251, 5.732933, 5.73209] * u.arcsec
    pa = [-32.942623, -32.931957, -33.07815, -33.07532, -33.187653,
          -33.175243, -33.167213, -33.167244, -33.170418, -33.180233] * u.deg

    beams = Beams(major=major, minor=minor, pa=pa)

    err_str = 'Could not find common beam to deconvolve all beams.'
    with pytest.raises(BeamError, match=err_str):

        com_beam = beams.common_beam(method='pts',
                                     epsilon=5e-4,
                                     auto_increase_epsilon=False)

    # Force running into the max iteration of epsilon increases.
    err_str = 'Could not increase epsilon to find common beam.'
    with pytest.raises(BeamError, match=err_str):

        com_beam = beams.common_beam(method='pts',
                                     epsilon=5e-4,
                                     max_iter=2,
                                     max_epsilon=6e-4,
                                     auto_increase_epsilon=True)

    # Should run when epsilon is allowed to increase a bit.
    com_beam = beams.common_beam(method='pts',
                                 epsilon=5e-4,
                                 auto_increase_epsilon=True,
                                 max_epsilon=1e-3)



import six
from astropy import units as u
from astropy.io import fits
from astropy import constants
from astropy import wcs
import numpy as np
import warnings

# Imports for the custom kernels
from astropy.modeling.models import Ellipse2D, Gaussian2D
from astropy.convolution import Kernel2D
from astropy.convolution.kernels import _round_up_to_odd_integer

from .utils import deconvolve_optimized, convolve, RadioBeamDeprecationWarning

# Conversion between a twod Gaussian FWHM**2 and effective area
FWHM_TO_AREA = 2*np.pi/(8*np.log(2))
SIGMA_TO_FWHM = np.sqrt(8*np.log(2))

class NoBeamException(Exception):
    pass

def _to_area(major,minor):
    return (major * minor * FWHM_TO_AREA).to(u.sr)

unit_format = {u.deg: r'\\circ',
               u.arcsec: "''",
               u.arcmin: "'"}


class Beam(u.Quantity):
    """
    An object to handle single radio beams.
    """

    def __new__(cls, major=None, minor=None, pa=None, area=None,
                default_unit=u.arcsec, meta=None):
        """
        Create a new Gaussian beam

        Parameters
        ----------
        major : :class:`~astropy.units.Quantity` with angular equivalency
            The FWHM major axis
        minor : :class:`~astropy.units.Quantity` with angular equivalency
            The FWHM minor axis
        pa : :class:`~astropy.units.Quantity` with angular equivalency
            The beam position angle
        area : :class:`~astropy.units.Quantity` with steradian equivalency
            The area of the beam.  This is an alternative to specifying the
            major/minor/PA, and will create those values assuming a circular
            Gaussian beam.
        default_unit : :class:`~astropy.units.Unit`
            The unit to impose on major, minor if they are specified as floats
        """

        # improve to some kwargs magic later

        # error checking

        # ... given an area make a round beam assuming it is Gaussian
        if area is not None:
            if major is not None:
                raise ValueError("Can only specify one of {major,minor,pa} "
                                 "and {area}")
            rad = np.sqrt(area/(2*np.pi)) * u.deg
            major = rad * SIGMA_TO_FWHM
            minor = rad * SIGMA_TO_FWHM
            pa = 0.0 * u.deg

        # give specified values priority
        if major is not None:
            if u.deg.is_equivalent(major):
                major = major
            else:
                warnings.warn("Assuming major axis has been specified in degrees")
                major = major * u.deg
        if minor is not None:
            if u.deg.is_equivalent(minor):
                minor = minor
            else:
                warnings.warn("Assuming minor axis has been specified in degrees")
                minor = minor * u.deg
        if pa is not None:
            if u.deg.is_equivalent(pa):
                pa = pa
            else:
                warnings.warn("Assuming position angle has been specified in degrees")
                pa = pa * u.deg
        else:
            pa = 0.0 * u.deg

        # some sensible defaults
        if minor is None:
            minor = major

        if minor > major:
            raise ValueError("Minor axis greater than major axis.")

        self = super(Beam, cls).__new__(cls, _to_area(major,minor).value, u.sr)
        self._major = major
        self._minor = minor
        self._pa = pa
        self.default_unit = default_unit

        if meta is None:
            self.meta = {}
        elif isinstance(meta, dict):
            self.meta = meta
        else:
            raise TypeError("metadata must be a dictionary")

        return self

    @classmethod
    def from_fits_bintable(cls, bintable, tolerance=0.01):
        """
        Instantiate a single beam from a bintable from a CASA-produced image
        HDU.  The beams in the BinTableHDU will be averaged to form a single
        beam.

        Parameters
        ----------
        bintable : fits.BinTableHDU
            The table data containing the beam information
        tolerance : float
            The fractional tolerance on the beam size to include when averaging
            to a single beam

        Returns
        -------
        beam : Beam
            A new beam object that is the average of the table beams
        """
        from astropy.stats import circmean

        bmaj = bintable.data['BMAJ']
        bmin = bintable.data['BMIN']
        bpa = bintable.data['BPA']
        if np.any(np.isnan(bmaj) | np.isnan(bmin) | np.isnan(bpa)):
            raise ValueError("NaN beam encountered.")
        for par in (bmin,bmaj):
            par_mean = par.mean()
            if (par.max() > par_mean*(1+tolerance)) or (par.min()<par_mean*(1-tolerance)):
                raise ValueError("Beams are not within specified tolerance")

        meta = {key: bintable.data[key].mean() for key in bintable.data.names if
                key not in ('BMAJ','BPA', 'BMIN')}
        if meta:
            warnings.warn("Metadata was averaged for keywords "
                          "{0}".format(",".join([key for key in meta])))

        return cls(major=bmaj.mean()*u.arcsec, minor=bmin.mean()*u.arcsec,
                   pa=circmean(bpa*u.deg, weights=bmaj/bmin))

    @classmethod
    def from_fits_header(cls, hdr):
        """
        Instantiate the beam from a header. Attempts to extract the
        beam from standard keywords. Failing that, it looks for an
        AIPS-style HISTORY entry.
        """
        # ... given a file try to make a fits header
        # assume a string refers to a filename on disk
        if not isinstance(hdr,fits.Header):
            if isinstance(hdr, six.string_types):
                if hdr.lower().endswith(('.fits', '.fits.gz', '.fit',
                                         '.fit.gz', '.fits.Z', '.fit.Z')):
                    hdr = fits.getheader(hdr)
                else:
                    raise TypeError("Unrecognized extension.")
            else:
                raise TypeError("Header is not a FITS header or a filename")


        # If we find a major axis keyword then we are in keyword
        # mode. Else look to see if there is an AIPS header.
        if "BMAJ" in hdr:
            major = hdr["BMAJ"] * u.deg
        else:
            hist_beam = cls.from_fits_history(hdr)
            if hist_beam is not None:
                return hist_beam
            else:
                raise NoBeamException("No BMAJ found and does not appear to be a CASA/AIPS header.")

        # Fill out the minor axis and position angle if they are
        # present. Else they will default .
        if "BMIN" in hdr:
            minor = hdr["BMIN"] * u.deg
        else:
            minor = None
        if "BPA" in hdr:
            pa = hdr["BPA"] * u.deg
        else:
            pa = None

        return cls(major=major, minor=minor, pa=pa)


    @classmethod
    def from_fits_history(cls, hdr):
        """
        Instantiate the beam from an AIPS header. AIPS holds the beam
        in history. This method of initializing uses the last such
        entry.
        """
        # a line looks like
        # HISTORY AIPS   CLEAN BMAJ=  1.7599E-03 BMIN=  1.5740E-03 BPA=   2.61
        if 'HISTORY' not in hdr:
            return None

        aipsline = None
        for line in hdr['HISTORY']:
            if 'BMAJ' in line:
                aipsline = line

        # a line looks like
        # HISTORY Sat May 10 20:53:11 2014
        # HISTORY imager::clean() [] Fitted beam used in
        # HISTORY > restoration: 1.34841 by 0.830715 (arcsec)
        #        at pa 82.8827 (deg)

        casaline = None
        for line in hdr['HISTORY']:
            if ('restoration' in line) and ('arcsec' in line):
                casaline = line
        #assert precedence for CASA style over AIPS
        #        this is a dubious choice

        if casaline is not None:
            bmaj = float(casaline.split()[2]) * u.arcsec
            bmin = float(casaline.split()[4]) * u.arcsec
            bpa = float(casaline.split()[8]) * u.deg
            return cls(major=bmaj, minor=bmin, pa=bpa)

        elif aipsline is not None:
            bmaj = float(aipsline.split()[3]) * u.deg
            bmin = float(aipsline.split()[5]) * u.deg
            bpa = float(aipsline.split()[7]) * u.deg
            return cls(major=bmaj, minor=bmin, pa=bpa)

        else:
            return None

    @classmethod
    def from_casa_image(cls, imagename):
        '''
        Instantiate beam from a CASA image.

        ** Must be run in a CASA environment! **

        Parameters
        ----------
        imagename : str
            Name of CASA image.
        '''

        try:
            import casac
        except ImportError:
            raise ImportError("Could not import CASA (casac) and therefore"
                              " cannot read CASA .image files")

        ia.open(imagename)
        beam_props = ia.restoringbeam()
        ia.close()

        beam_keys = ["major", "minor", "positionangle"]
        if not all([True for key in beam_keys if key in beam_props]):
            raise ValueError("The image does not contain complete beam "
                             "information. Check the output of "
                             "ia.restoringbeam().")

        major = beam_props["major"]["value"] * \
            u.Unit(beam_props["major"]["unit"])
        minor = beam_props["minor"]["value"] * \
            u.Unit(beam_props["minor"]["unit"])
        pa = beam_props["positionangle"]["value"] * \
            u.Unit(beam_props["positionangle"]["unit"])

        return cls(major=major, minor=minor, pa=pa)

    def attach_to_header(self, header, copy=True):
        '''
        Attach the beam information to the provided header.

        Parameters
        ----------
        header : astropy.io.fits.header.Header
            Header to add/update beam info.
        copy : bool, optional
            Returns a copy of the inputted header with the beam information.

        Returns
        -------
        copy_header : astropy.io.fits.header.Header
            Copy of the input header with the updated beam info when
            `copy=True`.
        '''

        if copy:
            header = header.copy()

        header.update(self.to_header_keywords())

        return header

    def __repr__(self):
        return "Beam: BMAJ={0} BMIN={1} BPA={2}".format(self.major.to(self.default_unit),self.minor.to(self.default_unit),self.pa.to(u.deg))

    def __repr_html__(self):
        return "Beam: BMAJ={0} BMIN={1} BPA={2}".format(self.major.to(self.default_unit),self.minor.to(self.default_unit),self.pa.to(u.deg))

    def _repr_latex_(self):
        return "Beam: BMAJ=${0}^{{{fmt}}}$ BMIN=${1}^{{{fmt}}}$ BPA=${2}^\\circ$".format(self.major.to(self.default_unit).value,
                                                                                         self.minor.to(self.default_unit).value,
                                                                                         self.pa.to(u.deg).value,
                                                                                         fmt = unit_format[self.default_unit])

    def __str__(self):
        return self.__repr__()


    def convolve(self, other):
        """
        Convolve one beam with another.

        Parameters
        ----------
        other : `Beam`
            The beam to convolve with

        Returns
        -------
        new_beam : `Beam`
            The convolved Beam
        """

        new_major, new_minor, new_pa = convolve(self, other)

        return Beam(major=new_major,
                    minor=new_minor,
                    pa=new_pa)

    def __mul__(self, other):
        return self.convolve(other)

    # Does division do the same? Or what? Doesn't have to be defined.
    def __sub__(self, other):
        warnings.warn("Subtraction-as-deconvolution is deprecated. "
                      "Use division instead.",
                      RadioBeamDeprecationWarning)
        return self.deconvolve(other)

    def __truediv__(self, other):
        return self.deconvolve(other)

    def deconvolve(self, other, failure_returns_pointlike=False):
        """
        Deconvolve a beam from another

        Parameters
        ----------
        other : `Beam`
            The beam to deconvolve from this beam
        failure_returns_pointlike : bool
            Option to return a pointlike beam (i.e., one with major=minor=0) if
            the second beam is larger than the first.  Otherwise, a ValueError
            will be raised

        Returns
        -------
        new_beam : `Beam`
            The convolved Beam

        Raises
        ------
        failure : ValueError
            If the second beam is larger than the first, the default behavior
            is to raise an exception.  This can be overridden with
            failure_returns_pointlike
        """

        new_major, new_minor, new_pa = deconvolve_optimized(self.to_header_keywords(),
                                                            other.to_header_keywords(),
                                                            failure_returns_pointlike=failure_returns_pointlike)

        # Keep the units from before
        new_major = (new_major * u.deg).to(self.major.unit)
        new_minor = (new_minor * u.deg).to(self.minor.unit)
        new_pa = (new_pa * u.rad).to(self.pa.unit)

        return Beam(major=new_major, minor=new_minor, pa=new_pa)


    def __eq__(self, other):

        # Catch floating point issues
        atol_deg = 1e-12 * u.deg

        this_pa = self.pa.to(u.deg) % (180.0 * u.deg)
        other_pa = other.pa.to(u.deg) % (180.0 * u.deg)

        if self.iscircular():
            equal_pa = True
        else:
            equal_pa = True if np.abs(this_pa - other_pa) < atol_deg else False

        equal_maj = np.abs(self.major - other.major) < atol_deg
        equal_min = np.abs(self.minor - other.minor) < atol_deg

        if equal_maj and equal_min and equal_pa:
            return True
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    # Is it astropy convention to access properties through methods?
    @property
    def sr(self):
        return _to_area(self.major,self.minor)

    @property
    def major(self):
        """ Beam FWHM Major Axis """
        return self._major

    @property
    def minor(self):
        """ Beam FWHM Minor Axis """
        return self._minor

    @property
    def pa(self):
        return self._pa

    @property
    def isfinite(self):
        return ((self.major > 0) & (self.minor > 0) & np.isfinite(self.major) &
                np.isfinite(self.minor) & np.isfinite(self.pa))

    def iscircular(self, rtol=1e-6):

        frac_diff = (self.major - self.minor).to(u.deg) / self.major.to(u.deg)

        return frac_diff <= rtol

    def beam_projected_area(self, distance):
        """
        Return the beam area in pc^2 (or equivalent) given a distance
        """
        return self.sr*(distance**2)/u.sr

    def jtok_equiv(self, freq):
        '''
        Return conversion function between Jy/beam to K at the specified
        frequency.

        The function can be used with the usual astropy.units conversion:
        >>> beam = Beam.from_fits_header("header.fits") # doctest: +SKIP
        >>> (1.0*u.Jy).to(u.K, beam.jtok_equiv(1.4*u.GHz)) # doctest: +SKIP

        Parameters
        ----------
        freq : astropy.units.quantity.Quantity
            Frequency to calculate conversion.

        Returns
        -------
        u.brightness_temperature
        '''

        if not isinstance(freq, u.quantity.Quantity):
            raise TypeError("freq must be a Quantity object. "
                            "Try 'freq*u.Hz' or another equivalent unit.")

        try:
            return u.brightness_temperature(beam_area=self.sr, frequency=freq)
        except TypeError:
            # old astropy used ordered arguments
            return u.brightness_temperature(self.sr, freq)

    def jtok(self, freq, value=1.0*u.Jy):
        """
        Return the conversion for the given value between Jy/beam to K at
        the specified frequency.

        Unlike :meth:`jtok_equiv`, the output is the numerical value that
        converts the units, without any attached unit.

        Parameters
        ----------
        freq : astropy.units.quantity.Quantity
            Frequency to calculate conversion.
        value : astropy.units.quantity.Quantity
            Value (in Jy or an equivalent unit) to convert to K.

        Returns
        -------
        value : float
            Value converted to K.
        """

        return value.to(u.K, self.jtok_equiv(freq))

    @property
    def beamarea_equiv(self):
        return u.beam_angular_area(self.sr)

    def ellipse_to_plot(self, xcen, ycen, pixscale):
        """
        Return a matplotlib ellipse for plotting

        Parameters
        ----------
        xcen : int
            Center pixel in the x-direction.
        ycen : int
            Center pixel in the y-direction.
        pixscale : `~astropy.units.Quantity`
            Conversion from degrees to pixels.

        Returns
        -------
        ~matplotlib.patches.Ellipse
            Ellipse patch object centered on the given pixel coordinates.
        """
        from matplotlib.patches import Ellipse
        return Ellipse((xcen, ycen),
                       width=(self.major.to(u.deg) / pixscale).to(u.dimensionless_unscaled).value,
                       height=(self.minor.to(u.deg) / pixscale).to(u.dimensionless_unscaled).value,
                       # PA is 90 deg offset from x-y axes by convention
                       # (it is angle from NCP)
                       angle=(self.pa+90*u.deg).to(u.deg).value)

    def as_kernel(self, pixscale, **kwargs):
        """
        Returns an elliptical Gaussian kernel of the beam.

        .. warning::
            This method is not aware of any misalignment between pixel
            and world coordinates.

        Parameters
        ----------
        pixscale : `~astropy.units.Quantity`
            Conversion from angular to pixel size.
        kwargs : passed to EllipticalGaussian2DKernel
        """
        # do something here involving matrices
        # need to rotate the kernel into the wcs pixel space, kinda...
        # at the least, need to rescale the kernel axes into pixels

        stddev_maj = (self.major.to(u.deg)/(pixscale.to(u.deg) *
                                            SIGMA_TO_FWHM)).decompose()
        stddev_min = (self.minor.to(u.deg)/(pixscale.to(u.deg) *
                                            SIGMA_TO_FWHM)).decompose()

        # position angle is defined as CCW from north
        # "angle" is conventionally defined as CCW from "west".
        # Therefore, add 90 degrees
        angle = (90*u.deg+self.pa).to(u.radian).value,

        return EllipticalGaussian2DKernel(stddev_maj.value,
                                          stddev_min.value,
                                          angle,
                                          **kwargs)

    def as_tophat_kernel(self, pixscale, **kwargs):
        '''
        Returns an elliptical Tophat kernel of the beam. The area has
        been scaled to match the 2D Gaussian area:

        .. math::
            \\begin{array}{ll}
            A_{\\mathrm{Gauss}} = 2\\pi\\sigma_{\\mathrm{Gauss}}^{2}
            A_{\\mathrm{Tophat}} = \\pi\\sigma_{\\mathrm{Tophat}}^{2}
            \\sigma_{\\mathrm{Tophat}} = \\sqrt{2}\\sigma_{\\mathrm{Gauss}}
            \\end{array}

        .. warning::
            This method is not aware of any misalignment between pixel
            and world coordinates.

        Parameters
        ----------
        pixscale : float
            deg -> pixels
        **kwargs : passed to EllipticalTophat2DKernel
        '''

        # Based on Gaussian to Tophat area conversion
        # A_gaussian = 2 * pi * sigma^2 / (sqrt(8*log(2))^2
        # A_tophat = pi * r^2
        # pi r^2 = 2 * pi * sigma^2 / (sqrt(8*log(2))^2
        # r = sqrt(2)/sqrt(8*log(2)) * sigma

        gauss_to_top = np.sqrt(2)

        maj_eff = gauss_to_top * self.major.to(u.deg) / \
            (pixscale * SIGMA_TO_FWHM)
        min_eff = gauss_to_top * self.minor.to(u.deg) / \
            (pixscale * SIGMA_TO_FWHM)

        # position angle is defined as CCW from north
        # "angle" is conventionally defined as CCW from "west".
        # Therefore, add 90 degrees
        angle = (90*u.deg+self.pa).to(u.radian).value,

        return EllipticalTophat2DKernel(maj_eff.value, min_eff.value,
                                        angle, **kwargs)

    def to_header_keywords(self):
        return {'BMAJ': self.major.to(u.deg).value,
                'BMIN': self.minor.to(u.deg).value,
                'BPA':  self.pa.to(u.deg).value,
                }

# Beam.__doc__ = Beam.__doc__ + Beam.__new__.__doc__

def mywcs_to_platescale(mywcs):
    pix_area = wcs.utils.proj_plane_pixel_area(mywcs)
    return pix_area**0.5


class EllipticalGaussian2DKernel(Kernel2D):
    """
    2D Elliptical Gaussian filter kernel.

    The Gaussian filter is a filter with great smoothing properties. It is
    isotropic and does not produce artifacts.

    Parameters
    ----------
    stddev_maj : float
        Standard deviation of the Gaussian kernel in direction 1
    stddev_min : float
        Standard deviation of the Gaussian kernel in direction 1
    position_angle : float
        Position angle of the elliptical gaussian
    x_size : odd int, optional
        Size in x direction of the kernel array. Default = support_scaling *
        stddev.
    y_size : odd int, optional
        Size in y direction of the kernel array. Default = support_scaling *
        stddev.
    support_scaling : int
        The amount to scale the stddev to determine the size of the kernel
    mode : str, optional
        One of the following discretization modes:
            * 'center' (default)
                Discretize model by taking the value
                at the center of the bin.
            * 'linear_interp'
                Discretize model by performing a bilinear interpolation
                between the values at the corners of the bin.
            * 'oversample'
                Discretize model by taking the average
                on an oversampled grid.
            * 'integrate'
                Discretize model by integrating the
                model over the bin.
    factor : number, optional
        Factor of oversampling. Default factor = 10.


    See Also
    --------
    Box2DKernel, Tophat2DKernel, MexicanHat2DKernel, Ring2DKernel,
    TrapezoidDisk2DKernel, AiryDisk2DKernel, Gaussian2DKernel,
    EllipticalTophat2DKernel

    Examples
    --------
    Kernel response:

     .. plot::
        :include-source:

        import matplotlib.pyplot as plt
        from radio_beam import EllipticalGaussian2DKernel
        gaussian_2D_kernel = EllipticalGaussian2DKernel(10, 5, np.pi/4)
        plt.imshow(gaussian_2D_kernel, interpolation='none', origin='lower')
        plt.xlabel('x [pixels]')
        plt.ylabel('y [pixels]')
        plt.colorbar()
        plt.show()

    """
    _separable = True
    _is_bool = False

    def __init__(self, stddev_maj, stddev_min, position_angle,
                 support_scaling=8, **kwargs):
        self._model = Gaussian2D(1. / (2 * np.pi * stddev_maj * stddev_min), 0,
                                 0, x_stddev=stddev_maj, y_stddev=stddev_min,
                                 theta=position_angle)

        try:
            from astropy.modeling.utils import ellipse_extent
        except ImportError:
            raise NotImplementedError("EllipticalGaussian2DKernel requires"
                                      " astropy 1.1b1 or greater.")

        max_extent = \
            np.max(ellipse_extent(stddev_maj, stddev_min, position_angle))
        self._default_size = \
            _round_up_to_odd_integer(support_scaling * 2 * max_extent)
        super(EllipticalGaussian2DKernel, self).__init__(**kwargs)
        self._truncation = np.abs(1. - 1 / self._array.sum())


class EllipticalTophat2DKernel(Kernel2D):
    """
    2D Elliptical Tophat filter kernel.

    The Tophat filter can produce artifacts when applied
    repeatedly on the same data.

    Parameters
    ----------
    stddev_maj : float
        Standard deviation of the Gaussian kernel in direction 1
    stddev_min : float
        Standard deviation of the Gaussian kernel in direction 1
    position_angle : float
        Position angle of the elliptical gaussian
    x_size : odd int, optional
        Size in x direction of the kernel array. Default = support_scaling *
        stddev.
    y_size : odd int, optional
        Size in y direction of the kernel array. Default = support_scaling *
        stddev.
    support_scaling : int
        The amount to scale the stddev to determine the size of the kernel
    mode : str, optional
        One of the following discretization modes:
            * 'center' (default)
                Discretize model by taking the value
                at the center of the bin.
            * 'linear_interp'
                Discretize model by performing a bilinear interpolation
                between the values at the corners of the bin.
            * 'oversample'
                Discretize model by taking the average
                on an oversampled grid.
            * 'integrate'
                Discretize model by integrating the
                model over the bin.
    factor : number, optional
        Factor of oversampling. Default factor = 10.


    See Also
    --------
    Box2DKernel, Tophat2DKernel, MexicanHat2DKernel, Ring2DKernel,
    TrapezoidDisk2DKernel, AiryDisk2DKernel, Gaussian2DKernel,
    EllipticalGaussian2DKernel

    Examples
    --------
    Kernel response:

     .. plot::
        :include-source:

        import matplotlib.pyplot as plt
        from radio_beam import EllipticalTophat2DKernel
        tophat_2D_kernel = EllipticalTophat2DKernel(10, 5, np.pi/4)
        plt.imshow(tophat_2D_kernel, interpolation='none', origin='lower')
        plt.xlabel('x [pixels]')
        plt.ylabel('y [pixels]')
        plt.colorbar()
        plt.show()

    """

    _is_bool = True

    def __init__(self, stddev_maj, stddev_min, position_angle, support_scaling=1,
                 **kwargs):

        self._model = Ellipse2D(1. / (np.pi * stddev_maj * stddev_min), 0, 0,
                                stddev_maj, stddev_min, position_angle)

        try:
            from astropy.modeling.utils import ellipse_extent
        except ImportError:
            raise NotImplementedError("EllipticalTophat2DKernel requires"
                                      " astropy 1.1b1 or greater.")

        max_extent = \
            np.max(ellipse_extent(stddev_maj, stddev_min, position_angle))
        self._default_size = \
            _round_up_to_odd_integer(support_scaling * 2 * max_extent)
        super(EllipticalTophat2DKernel, self).__init__(**kwargs)
        self._truncation = 0




'''

UNUSED IN CURRENT TESTING SUITE!

Run in a CASA environment to generate the smallest common beam for a range of
parameters.
'''

RUN = False

if RUN:

    try:
        import casac
    except ImportError:
        raise ImportError("This script must be run in CASA")

    # Tested for CASA 4.7.2
    import casadef
    if casadef.casa_version != "4.7.2":
        raise Exception("This script is only tested in CASA 4.7.2. "
                        "Found version {}".format(casadef.casa_version))

    import numpy as np

    # Generate fake cube w/ 2 spectral channels
    im1 = ia.newimagefromarray(pixels=ia.makearray(0, [4, 4, 1, 2]))

    # Give the first channel a circular beam
    im1.setrestoringbeam(major='3arcsec', minor='3arcsec', pa='0deg', channel=0)


    # Give the other channel an elongated beam, and rotate it
    majors = []
    minors = []
    pas = []

    # Include the PA of the one beam so tests can be run later.
    orig_pas = np.arange(179)

    for pa in orig_pas:
        im1.setrestoringbeam(major='4arcsec', minor='2.5arcsec',
                             pa='{}deg'.format(pa), channel=1)

        com_beam = im1.commonbeam()

        majors.append(com_beam['major']['value'])
        minors.append(com_beam['minor']['value'])
        pas.append(com_beam['pa']['value'])

    common_beams = np.array([orig_pas, majors, minors, pas]).T
    np.savetxt("commonbeam_CASA_comparison.csv", common_beams, delimiter=",")

    # im1 = ia.newimagefromarray(pixels=ia.makearray(0, [4, 4, 1, 4]))
    # for i, pa in enumerate([0, 20, 40, 60]):
    #     im1.setrestoringbeam(channel=i, major='4arcsec', minor='3arcsec',
    #                          pa="{}deg".format(pa))

    # im2 = ia.newimagefromarray(pixels=ia.makearray(0, [4, 4, 1, 4]))
    # for i, pa in enumerate([0, 60, 20, 40]):
    #     im2.setrestoringbeam(channel=i, major='4arcsec', minor='3arcsec',
    #                          pa="{}deg".format(pa))

    # im3 = ia.newimagefromarray(pixels=ia.makearray(0, [4, 4, 1, 4]))
    # for i, pa in enumerate([60, 40, 20, 0]):
    #     im3.setrestoringbeam(channel=i, major='4arcsec', minor='3arcsec',
    #                          pa="{}deg".format(pa))

    # im4 = ia.newimagefromarray(pixels=ia.makearray(0, [4, 4, 1, 4]))
    # for i, pa in enumerate([60, 0, 20, 40]):
    #     im4.setrestoringbeam(channel=i, major='4arcsec', minor='3arcsec',
    #                          pa="{}deg".format(pa))

    # print(im1.commonbeam())
    # print(im2.commonbeam())
    # print(im3.commonbeam())
    # print(im4.commonbeam())




from astropy import units as u
from astropy.io import fits
from astropy import constants
from astropy import wcs
import numpy as np
import warnings

from .beam import Beam, _to_area, SIGMA_TO_FWHM
from .commonbeam import commonbeam
from .utils import InvalidBeamOperationError


class Beams(u.Quantity):
    """
    An object to handle a set of radio beams for a data cube.
    """

    def __new__(cls, major=None, minor=None, pa=None,
                areas=None, default_unit=u.arcsec, meta=None,
                beams=None):
        """
        Create a new set of Gaussian beams

        Parameters
        ----------
        major : :class:`~astropy.units.Quantity` with angular equivalency
            The FWHM major axes
        minor : :class:`~astropy.units.Quantity` with angular equivalency
            The FWHM minor axes
        pa : :class:`~astropy.units.Quantity` with angular equivalency
            The beam position angles
        area : :class:`~astropy.units.Quantity` with steradian equivalency
            The area of the beams.  This is an alternative to specifying the
            major/minor/PA, and will create those values assuming a circular
            Gaussian beam.
        default_unit : :class:`~astropy.units.Unit`
            The unit to impose on major, minor if they are specified as floats
        beams : List of :class:`~radio_beam.Beam` objects
            List of individual `Beam` objects. The resulting `Beams` object will
            have major and minor axes in degrees.
        """

        # improve to some kwargs magic later

        # error checking

        if beams is not None:
            major = [beam.major.to(u.deg).value for beam in beams] * u.deg
            minor = [beam.minor.to(u.deg).value for beam in beams] * u.deg
            pa = [beam.pa.to(u.deg).value for beam in beams] * u.deg

        # ... given an area make a round beam assuming it is Gaussian
        if areas is not None:
            rad = np.sqrt(areas / (2 * np.pi)) * u.deg
            major = rad * SIGMA_TO_FWHM
            minor = rad * SIGMA_TO_FWHM
            pa = np.zeros_like(areas) * u.deg

        # give specified values priority
        if major is not None:
            if u.deg.is_equivalent(major.unit):
                pass
            else:
                warnings.warn("Assuming major axes has been specified in degrees")
                major = major * u.deg
        if minor is not None:
            if u.deg.is_equivalent(minor.unit):
                pass
            else:
                warnings.warn("Assuming minor axes has been specified in degrees")
                minor = minor * u.deg
        if pa is not None:
            if len(pa) != len(major):
                raise ValueError("Number of position angles must match number of major axis lengths")
            if u.deg.is_equivalent(pa.unit):
                pass
            else:
                warnings.warn("Assuming position angles has been specified in degrees")
                pa = pa * u.deg
        else:
            pa = np.zeros_like(major.value) * u.deg

        # some sensible defaults
        if minor is None:
            minor = major
        elif len(minor) != len(major):
            raise ValueError("Minor and major axes must have same number of values")

        if np.any(minor > major):
            raise ValueError("Minor axis greater than major axis.")

        self = super(Beams, cls).__new__(cls, value=_to_area(major, minor).value, unit=u.sr)
        self.major = major
        self.minor = minor
        self.pa = pa
        self.default_unit = default_unit

        if meta is None:
            self.meta = [{}]*len(self)
        else:
            self.meta = meta

        return self

    @property
    def meta(self):
        return self._meta

    @meta.setter
    def meta(self, value):
        if len(value) == len(self):
            self._meta = value
        else:
            raise TypeError("metadata must be a list of dictionaries")

    def __len__(self):
        return len(self.major)


    @property
    def isfinite(self):
        return ((self.major > 0) & (self.minor > 0) & np.isfinite(self.major) &
                np.isfinite(self.minor) & np.isfinite(self.pa))

    def __getslice__(self, start, stop, increment=None):
        return self.__getitem__(slice(start, stop, increment))

    def __getitem__(self, view):
        if isinstance(view, (int, np.int64)):
            return Beam(major=self.major[view],
                        minor=self.minor[view],
                        pa=self.pa[view],
                        meta=self.meta[view])
        elif isinstance(view, slice):
            return Beams(major=self.major[view],
                         minor=self.minor[view],
                         pa=self.pa[view],
                         meta=self.meta[view])
        elif isinstance(view, np.ndarray):
            if view.dtype.name != 'bool':
                raise ValueError("If using an array to index beams, it must "
                                 "be a boolean array.")
            return Beams(major=self.major[view],
                         minor=self.minor[view],
                         pa=self.pa[view],
                         meta=[x for ii,x in zip(view, self.meta) if ii])
        else:
            raise ValueError("Invalid slice")

    def __array_finalize__(self, obj):
        # If our unit is not set and obj has a valid one, use it.
        if self._unit is None:
            unit = getattr(obj, '_unit', None)
            if unit is not None:
                self._set_unit(unit)

        if isinstance(obj, Beams):
            # Multiplication and division should change the area,
            # but not the PA or major/minor ratio
            self.major = obj.major
            self.minor = obj.minor
            self.pa = obj.pa
            self.meta = obj.meta

        # Copy info if the original had `info` defined.  Because of the way the
        # DataInfo works, `'info' in obj.__dict__` is False until the
        # `info` attribute is accessed or set.  Note that `obj` can be an
        # ndarray which doesn't have a `__dict__`.
        if 'info' in getattr(obj, '__dict__', ()):
            self.info = obj.info

    @property
    def sr(self):
        return _to_area(self.major, self.minor)

    @classmethod
    def from_fits_bintable(cls, bintable):
        """
        Instantiate a Beams list from a bintable from a CASA-produced image
        HDU.

        Parameters
        ----------
        bintable : fits.BinTableHDU
            The table data containing the beam information

        Returns
        -------
        beams : Beams
            A new Beams object
        """
        major = u.Quantity(bintable.data['BMAJ'], u.arcsec)
        minor = u.Quantity(bintable.data['BMIN'], u.arcsec)
        pa = u.Quantity(bintable.data['BPA'], u.deg)
        meta = [{key: row[key] for key in bintable.columns.names
                 if key not in ('BMAJ', 'BPA', 'BMIN')}
                for row in bintable.data]

        return cls(major=major, minor=minor, pa=pa, meta=meta)

    @classmethod
    def from_casa_image(cls, imagename):
        '''
        Instantiate beams from a CASA image. Cannot currently handle beams for
        different polarizations.

        ** Must be run in a CASA environment! **

        Parameters
        ----------
        imagename : str
            Name of CASA image.
        '''

        try:
            from casatools import image as iatool

            ia = iatool()

        except ImportError:

            raise ImportError("Could not import CASA and therefore"
                              " cannot read CASA .image files")

        ia.open(imagename)
        beam_props = ia.restoringbeam()
        ia.close()

        nchans = beam_props['nChannels']

        # Assuming there is always a 0th channel...
        maj_unit = u.Unit(beam_props['beams']['*0']['*0']['major']['unit'])
        min_unit = u.Unit(beam_props['beams']['*0']['*0']['minor']['unit'])
        pa_unit = u.Unit(beam_props['beams']['*0']['*0']['positionangle']['unit'])

        major = np.empty((nchans)) * maj_unit
        minor = np.empty((nchans)) * min_unit
        pa = np.empty((nchans)) * pa_unit

        for chan in range(nchans):

            chan_name = '*{}'.format(chan)
            chanbeam_props = beam_props['beams'][chan_name]['*0']

            # Can CASA have mixes of units between channels? Let's test just
            # in case
            assert maj_unit == u.Unit(chanbeam_props['major']['unit'])
            assert min_unit == u.Unit(chanbeam_props['minor']['unit'])
            assert pa_unit == u.Unit(chanbeam_props['positionangle']['unit'])

            major[chan] = chanbeam_props['major']['value'] * maj_unit
            minor[chan] = chanbeam_props['minor']['value'] * min_unit
            pa[chan] = chanbeam_props['positionangle']['value'] * pa_unit

        return cls(major=major, minor=minor, pa=pa)

    def average_beam(self, includemask=None, raise_for_nan=True):
        """
        Average the beam major, minor, and PA attributes.

        This is usually a dumb thing to do!
        """

        warnings.warn("Do not use the average beam for convolution! Use the"
                      " smallest common beam from `Beams.common_beam()`.")

        from astropy.stats import circmean

        if includemask is None:
            includemask = self.isfinite
        else:
            includemask = np.logical_and(includemask, self.isfinite)

        new_beam = Beam(major=self.major[includemask].mean(),
                        minor=self.minor[includemask].mean(),
                        pa=circmean(self.pa[includemask],
                        weights=(self.major / self.minor)[includemask]))

        if raise_for_nan and np.any(np.isnan(new_beam)):
            raise ValueError("NaNs after averaging.  This is a bug.")

        return new_beam

    def largest_beam(self, includemask=None):
        """
        Returns the largest beam (by area) in a list of beams.
        """

        if includemask is None:
            includemask = self.isfinite
        else:
            includemask = np.logical_and(includemask, self.isfinite)

        largest_idx = (self.major * self.minor)[includemask].argmax()
        new_beam = Beam(major=self.major[includemask][largest_idx],
                        minor=self.minor[includemask][largest_idx],
                        pa=self.pa[includemask][largest_idx])

        return new_beam

    def smallest_beam(self, includemask=None):
        """
        Returns the smallest beam (by area) in a list of beams.
        """

        if includemask is None:
            includemask = self.isfinite
        else:
            includemask = np.logical_and(includemask, self.isfinite)

        largest_idx = (self.major * self.minor)[includemask].argmin()
        new_beam = Beam(major=self.major[includemask][largest_idx],
                        minor=self.minor[includemask][largest_idx],
                        pa=self.pa[includemask][largest_idx])

        return new_beam

    def extrema_beams(self, includemask=None):
        return [self.smallest_beam(includemask),
                self.largest_beam(includemask)]

    def common_beam(self, includemask=None, method='pts', **kwargs):
        '''
        Return the smallest common beam size. For set of two beams,
        the solution is solved analytically. All larger sets solve for the
        minimum volume ellipse using the
        `Khachiyan Algorithm <http://www.mathworks.com/matlabcentral/fileexchange/9542>`_,
        where the convex hull of the set of ellipse edges is used to find the
        boundaries of the set.

        Since the minimum ellipse method is approximate, some solutions for
        the common beam will be slightly underestimated and the solution
        cannot be deconvolved from the whole set of beams. To overcome
        this issue, a small `epsilon` correction factor is added to the
        ellipse edges to encourage a valid common beam solution.
        Since `epsilon` is added to all sides, this correction will at most
        increase the common beam size by :math:`2\times(1+\epsilon)`.
        The default values of `epsilon` is :math:`5\times10^{-4}`, so this
        will have a very small effect on the size of the common beam.

        In some cases, `epsilon` must be increased to find a valid common
        beam solution. The algorithm does this by default
        (set by `auto_increase_epsilon=True`), and will incrementally
        increase `epsilon` until the common beam can be deconvolved from
        all beams, or until either(1) `max_iter` is reached (default is 10)
        or (2) `max_epsilon` is reached (default is 1e-3). In practice, we
        find these settings work well for different ALMA or VLA data, but
        these `kwargs` may need to be changed for discrepant cases.


        Parameters
        ----------
        includemask : `~numpy.ndarray`, optional
            Boolean mask.
        method : {'pts'}, optional
            Many beam method. Only `pts` is currently available.
        kwargs : Passed to `~radio_beam.commonbeam`.

        '''
        return commonbeam(self if includemask is None else self[includemask],
                          method=method, **kwargs)

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def __mul__(self, other):
        # Other must be a single beam. Assume multiplying is convolving
        # as set of beams with a given beam
        if not isinstance(other, Beam):
            raise InvalidBeamOperationError("Multiplication is defined as a "
                                            "convolution of the set of beams "
                                            "with a given beam. Must be "
                                            "multiplied with a Beam object.")

        return Beams(beams=[beam * other for beam in self])

    def __truediv__(self, other):
        # Other must be a single beam. Assume dividing is deconvolving
        # as set of beams with a given beam
        if not isinstance(other, Beam):
            raise InvalidBeamOperationError("Division is defined as a "
                                            "deconvolution of the set of beams"
                                            " with a given beam. Must be "
                                            "divided by a Beam object.")

        return Beams(beams=[beam / other for beam in self])

    def __add__(self, other):
        raise InvalidBeamOperationError("Addition of a set of Beams "
                                        "is not defined.")

    def __sub__(self, other):
        raise InvalidBeamOperationError("Addition of a set of Beams "
                                        "is not defined.")

    def __eq__(self, other):
        # other should be a single beam, or a another Beams object
        if isinstance(other, Beam):
            return np.array([beam == other for beam in self])
        elif isinstance(other, Beams):
            # These should have the same size.
            if not self.size == other.size:
                raise InvalidBeamOperationError("Beams objects must have the "
                                                "same shape to test "
                                                "equality.")

            return np.all([beam == other_beam for beam, other_beam in
                           zip(self, other)])
        else:
            raise InvalidBeamOperationError("Must test equality with a Beam"
                                            " or Beams object.")

    def __ne__(self, other):
        eq_out = self.__eq__(other)

        # If other is a Beam, will get array back
        if isinstance(eq_out, np.ndarray):
            return ~eq_out
        # If other is a Beams, will get boolean back
        else:
            return not eq_out




import numpy as np
import astropy.units as u

try:
    from scipy import optimize as opt
    from scipy.spatial import ConvexHull
    HAS_SCIPY = True
except ImportError:
    HAS_SCIPY = False

from .beam import Beam
from .utils import BeamError, transform_ellipse, deconvolve_optimized

__all__ = ['commonbeam', 'common_2beams', 'getMinVolEllipse',
           'common_manybeams_mve']


def commonbeam(beams, method='pts', **method_kwargs):
    '''
    Use analytic method if there are only two beams. Otherwise use constrained
    optimization to find the common beam.
    '''

    if beams.size == 1:
        return beams[0]
    elif fits_in_largest(beams):
        return beams.largest_beam()
    else:
        if beams.size == 2:
            try:
                return common_2beams(beams)
            # Sometimes this method can fail. Use the many beam solution in
            # this case
            except (ValueError, BeamError):
                pass

        if method == 'pts':
            return common_manybeams_mve(beams, **method_kwargs)
        elif method == 'opt':
            return common_manybeams_opt(beams, **method_kwargs)
        else:
            raise ValueError("method must be 'pts' or 'opt'.")


def common_2beams(beams, check_deconvolution=True):
    '''
    Find a common beam from a `Beams` object with 2 beams. This
    function is based on the CASA implementation `ia.commonbeam`. Note that
    the solution is only valid for 2 beams.

    Parameters
    ----------
    beams : `~radio_beam.Beams`
        Beams object with 2 beams.

    Returns
    -------
    common_beam : `~radio_beam.Beam`
        The smallest common beam in the set of beams.
    '''

    # This code is based on the implementation in CASA:
    # https://open-bitbucket.nrao.edu/projects/CASA/repos/casa/browse/code/imageanalysis/ImageAnalysis/CasaImageBeamSet.cc

    if beams.size != 2:
        raise BeamError("This method is only valid for two beams.")

    if (~beams.isfinite).all():
        raise BeamError("All beams in the object are invalid.")

    large_beam = beams.largest_beam()
    large_major = large_beam.major.to(u.arcsec)
    large_minor = large_beam.minor.to(u.arcsec)

    if beams.argmax() == 0:
        small_beam = beams[1]
    else:
        small_beam = beams[0]
    small_major = small_beam.major.to(u.arcsec)
    small_minor = small_beam.minor.to(u.arcsec)

    # Case where they're already equal
    if small_beam == large_beam:
        return large_beam

    deconv_beam = large_beam.deconvolve(small_beam, failure_returns_pointlike=True)

    # Larger beam can be deconvolved. It is already the smallest common beam
    if deconv_beam.isfinite:
        return large_beam

    # If the smaller beam is a circle, the minor axis is the circle radius
    if small_beam.iscircular():
        common_beam = Beam(large_beam.major, small_beam.major, large_beam.pa)
        return common_beam

    # Wrap angle about 0 to pi.
    pa_diff = ((small_beam.pa.to(u.rad).value - large_beam.pa.to(u.rad).value +
                np.pi / 2. + np.pi) % np.pi - np.pi / 2.) * u.rad

    # If the difference is pi / 2, the larger major is set to the
    # new major and the minor is the other major.
    if np.isclose(np.abs(pa_diff).value, np.pi / 2.):
        larger_major = large_beam.major >= small_beam.major
        major = large_major if larger_major else small_major
        minor = small_major if larger_major else small_major
        pa = large_beam.pa if larger_major else small_beam.pa
        conv_beam = Beam(major=major, minor=minor, pa=pa)

        return conv_beam

    else:
        # Transform to coordinates where large_beam is circular

        major_comb = np.sqrt(large_major * small_major)
        p = major_comb / large_major
        q = major_comb / large_minor

        # Transform beam into the same coordinates, and rotate so its
        # major axis is along the x axis.

        trans_major_sc, trans_minor_sc, trans_pa_sc = \
            transform_ellipse(small_major, small_minor, pa_diff, p, q)

        # The transformed minor axis is major_comb, as defined in CASA
        trans_minor_sc = major_comb

        # Return beam to the original coordinates, still rotated with
        # the major along the x axis
        trans_major_unsc, trans_minor_unsc, trans_pa_unsc = \
            transform_ellipse(trans_major_sc, trans_minor_sc,
                              trans_pa_sc, 1 / p, 1 / q)

        # Lastly, rotate the PA to the enclosing ellipse
        trans_major = trans_major_unsc.to(u.arcsec)
        trans_minor = trans_minor_unsc.to(u.arcsec)
        trans_pa = trans_pa_unsc + large_beam.pa

        # The minor axis becomes an issue when checking against the smaller
        # beam from deconvolution. Adding a tiny fraction makes the deconvolved
        # beam JUST larger than zero (~1e-7).
        epsilon = 100 * np.finfo(trans_major.dtype).eps * trans_major.unit
        trans_beam = Beam(major=trans_major + epsilon,
                          minor=trans_minor + epsilon,
                          pa=trans_pa)

        if check_deconvolution:
            # Ensure this beam can now be deconvolved
            deconv_large_beam = \
                trans_beam.deconvolve(large_beam,
                                    failure_returns_pointlike=True)
            deconv_prob_beam = \
                trans_beam.deconvolve(small_beam,
                                    failure_returns_pointlike=True)
            if not deconv_large_beam.isfinite or not deconv_prob_beam.isfinite:
                raise BeamError("Failed to find common beam that both beams can "
                                "be deconvolved by.")

        # Taken from CASA implementation, but by adding epsilon, this shouldn't
        # be needed

        # Scale the enclosing ellipse by a small factor until it does.
        # can_deconv = False
        # num = 0
        # while not can_deconv:

        #     deconv_large_beam = \
        #         trans_beam.deconvolve(large_beam,
        #                               failure_returns_pointlike=True)
        #     deconv_prob_beam = \
        #         trans_beam.deconvolve(small_beam,
        #                               failure_returns_pointlike=True)

        #     if (not deconv_large_beam.isfinite or not deconv_prob_beam.isfinite):

        #         scale_factor = 1.001
        #         trans_beam = Beam(major=trans_major * scale_factor,
        #                           minor=trans_minor * scale_factor,
        #                           pa=trans_pa)
        #     else:
        #         can_deconv = True

        #     if num == 10:
        #         break

        #     num += 1

        common_beam = trans_beam
        return common_beam


def boundingcircle(bmaj, bmin, bpa):
    thisone = np.argmax(bmaj)
    # PA really shouldn't matter here. But the minimization performed better
    # in some cases with a non-zero PA. Presumably this is b/c the PA of the
    # common beam is affected more by the beam with the largest major axis.
    return bmaj[thisone], bmaj[thisone], bpa[thisone]


def PtoA(bmaj, bmin, bpa):
    '''
    Express the ellipse parameters into
    `center-form <https://en.wikipedia.org/wiki/Matrix_representation_of_conic_sections>`_.
    '''
    A = np.zeros((2, 2))
    A[0, 0] = np.cos(bpa)**2 / bmaj**2 + np.sin(bpa)**2 / bmin**2
    A[1, 0] = np.cos(bpa) * np.sin(bpa) * (1 / bmaj**2 - 1 / bmin**2)
    A[0, 1] = A[1, 0]
    A[1, 1] = np.sin(bpa)**2 / bmaj**2 + np.cos(bpa)**2 / bmin**2
    return A


def BinsideA(B, A):
    try:
        np.linalg.cholesky(B - A)
        return True
    except np.linalg.LinAlgError:
        return False


def myobjective_regularized(p, bmajvec, bminvec, bpavec):
    # Force bmaj > bmin
    if p[0] < p[1]:
        return 1e30
    # We can safely assume the common major axis is at most the
    # largest major axis in the set
    if (p[0] <= bmajvec).any():
        return 1e30
    A = PtoA(*p)
    test = np.zeros_like(bmajvec)
    for idx, (bmx, bmn, bp) in enumerate(zip(bmajvec, bminvec, bpavec)):
        test[idx] = BinsideA(PtoA(bmx, bmn, bp), A)
    obj = 1 / np.linalg.det(A)
    if np.all(test):
        return obj
    else:
        return obj * 1e30


def common_manybeams_opt(beams, p0=None, opt_method='Nelder-Mead',
                         optdict={'maxiter': 5000, 'ftol': 1e-14,
                                  'maxfev': 5000},
                         verbose=False,
                         brute=False, brute_steps=40):
    '''
    Optimize the common beam solution by maximizing the determinant of the
    common beam.

    ..note:: This method is experimental and requires further testing.

    Parameters
    ----------
    beams : `~radio_beam.Beams`
        Beams object.
    p0 : tuple, optional
        Initial guess parameters (`major, minor, pa`).
    opt_method : str, optional
        Optimization method to use. See `~scipy.optimize.minimize`.
        The default of Nelder-Mead is the only method we have had
        some success with.
    optdict : dict, optional
        Dictionary parameters passed to `~scipy.optimize.minimize`.
    verbose : bool, optional
        Print the full output from `~scipy.optimize.minimize`.
    brute : bool, optional
        Use `~scipy.optimize.brute` to find the optimal solution.
    brute_steps : int, optional
        Number of positions to sample in each parameter (3).

    Returns
    -------
    com_beam : `~radio_beam.Beam`
        Common beam.
    '''

    raise NotImplementedError("This method is not fully tested. Remove this "
                              "line for testing purposes.")

    if not HAS_SCIPY:
        raise ImportError("common_manybeams_opt requires scipy.optimize.")

    bmaj = beams.major.value
    bmin = beams.minor.value
    bpa = beams.pa.to(u.rad).value

    if p0 is None:
        p0 = boundingcircle(bmaj, bmin, bpa)
        # It seems to help to make the initial guess slightly larger
        p0 = (1.1 * p0[0], 1.1 * p0[1], p0[2])

    if brute:
        maj_range = [beams.major.max(), 1.5 * beams.major.max()]
        maj_step = (maj_range[1] - maj_range[0]) / brute_steps
        min_range = [beams.minor.min(), 1.5 * beams.major.max()]
        min_step = (min_range[1] - min_range[0]) / brute_steps
        rranges = (slice(maj_range[0], maj_range[1], maj_step),
                   slice(min_range[0], min_range[1], min_step),
                   slice(0, 179.9, 180. / brute_steps))
        result = opt.brute(myobjective_regularized, rranges,
                           args=(bmaj, bmin, bpa),
                           full_output=True,
                           finish=opt.fmin)
        params = result[0]

    else:
        result = opt.minimize(myobjective_regularized, p0,
                              method=opt_method,
                              args=(bmaj, bmin, bpa),
                              options=optdict,
                              tol=1e-14)
        params = result.x

        if verbose:
            print(result.viewitems())

        if not result.success:
            raise Warning("Optimization failed")

    com_beam = Beam(params[0] * beams.major.unit,
                    params[1] * beams.major.unit,
                    (params[2] % np.pi) * u.rad)

    # Test if it deconvolves all
    if not fits_in_largest(beams, com_beam):
        raise BeamError("Could not find common beam to deconvolve all beams.")

    return com_beam


def fits_in_largest(beams, large_beam=None):
    '''
    Test if all beams can be deconvolved by the largest beam
    '''

    if large_beam is None:
        large_beam = beams.largest_beam()

    large_hdr_keywords = large_beam.to_header_keywords()

    majors = beams.major.to(u.deg).value
    minors = beams.minor.to(u.deg).value
    pas = beams.pa.to(u.deg).value

    # Catch differences below  << 1 microarsec = 2.8e10
    # This is the same limit used for checking equal beams in Beam.__eq__
    atol_limit = 1e-12

    for major, minor, pa in zip(majors, minors, pas):

        equal = abs(large_hdr_keywords['BMAJ'] - major) < atol_limit
        equal = equal and (abs(large_hdr_keywords['BMIN'] - minor) < atol_limit)

        # Check if the beam is circular
        # This checks for fractional changes below 1e-6 between the major and minor.
        # Same limit used in Beam.__eq__
        iscircular = (major - minor) / minor < 1e-6

        # position angle only matters if the beam is asymmetric
        if not iscircular:
            equal = equal and (abs(((large_hdr_keywords['BPA'] % np.pi) - (pa % np.pi))) < atol_limit)

        if equal:
            continue

        out = deconvolve_optimized(large_hdr_keywords,
                                   {'BMAJ': major,
                                    'BMIN': minor,
                                    'BPA': pa},
                                   failure_returns_pointlike=True)

        if np.any([ax == 0. for ax in out[:2]]):
            return False

    return True


def getMinVolEllipse(P, tolerance=1e-5, maxiter=1e5):
    """
    Use the Khachiyan Algorithm to compute that minimum volume ellipsoid.

    For the purposes of finding a common beam, there is an added check that
    requires the center to be within the tolerance range.

    Adapted code from: https://github.com/minillinim/ellipsoid/blob/master/ellipsoid.py

    That implementation relies on the original work by Nima Moshtagh:
    http://www.mathworks.com/matlabcentral/fileexchange/9542
    and an alternate python version from:
    http://cctbx.sourceforge.net/current/python/scitbx.math.minimum_covering_ellipsoid.html

    Parameters
    ----------
    P : `~numpy.ndarray`
        Points to compute solution.
    tolerance : float, optional
        Allowed error range in the Khachiyan Algorithm. Decreasing the
        tolerance by an order of magnitude requires an order of magnitude
        more iterations to converge.
    maxiter : int, optional
        Maximum iterations.

    Returns
    -------
    center : `~numpy.ndarray`
        Center point of the ellipse. Is required to be smaller than the
        tolerance.
    radii : `~numpy.ndarray`
        Radii of the ellipse.
    rotation : `~numpy.ndarray`
        Rotation matrix of the ellipse.

    """
    N, d = P.shape
    d = float(d)

    # Q will be our working array
    Q = np.vstack([np.copy(P.T), np.ones(N)])
    QT = Q.T

    # initializations
    err = 1.0
    u = np.ones(N) / N

    # Khachiyan Algorithm
    i = 0
    while err > tolerance:
        V = np.dot(Q, np.dot(np.diag(u), QT))
        # M the diagonal vector of an NxN matrix
        M = np.diag(np.dot(QT, np.dot(np.linalg.inv(V), Q)))
        j = np.argmax(M)
        maximum = M[j]
        step_size = (maximum - d - 1.0) / ((d + 1.0) * (maximum - 1.0))
        new_u = (1.0 - step_size) * u
        err = np.linalg.norm(new_u - u)
        if err <= tolerance:
            break
        new_u[j] += step_size
        u = new_u
        i += 1
        if i == maxiter:
            raise ValueError("Reached maximum iterations without converging."
                             " Try increasing the tolerance.")

    # center of the ellipse
    center = np.atleast_2d(np.dot(P.T, u))

    # For our purposes, the centre should always be very small
    center_square = np.outer(center, center)
    if not (center_square < tolerance**2).any():
        raise ValueError("The solved centre ({0}) is larger than the tolerance"
                         " ({1}). Check the input data.".format(center,
                                                                tolerance))

    # the A matrix for the ellipse
    A = np.linalg.inv(np.dot(P.T, np.dot(np.diag(u), P)) -
                      center_square) / d

    # ellip_vals = np.dot(P - center, np.dot(A, (P - center).T))
    # assert (ellip_vals <= 1. + tolerance).all()

    # Get the values we'd like to return
    U, s, rotation = np.linalg.svd(A)
    radii = 1.0 / np.sqrt(s)
    radii *= 1. + tolerance

    return center, radii, rotation


def ellipse_edges(beam, npts=300, epsilon=1e-3):
    '''
    Return the edge points of the beam.

    Parameters
    ----------
    beam : `~radio_beam.Beam`
        Beam object.
    npts : int, optional
        Number of samples.
    epsilon : float
        Increase the radii of the ellipse by 1 + epsilon. This is to ensure
        that `getMinVolEllipse` returns a marginally deconvolvable beam to
        within the error tolerance.

    Returns
    -------
    pts : `~numpy.ndarray`
        The x, y coordinates of the ellipse edge.
    '''

    bpa = beam.pa.to(u.rad).value
    major = beam.major.to(u.deg).value * (1. + epsilon)
    minor = beam.minor.to(u.deg).value * (1. + epsilon)

    phi = np.linspace(0, 2 * np.pi, npts)

    x = major * np.cos(phi)
    y = minor * np.sin(phi)

    xr = x * np.cos(bpa) - y * np.sin(bpa)
    yr = x * np.sin(bpa) + y * np.cos(bpa)

    pts = np.vstack([xr, yr])
    return pts


def common_manybeams_mve(beams, tolerance=1e-4, nsamps=200,
                         epsilon=5e-4,
                         auto_increase_epsilon=True,
                         max_epsilon=1e-3,
                         max_iter=10):
    '''
    Calculate a common beam size using the Khachiyan Algorithm to find the
    minimum enclosing ellipse from all beam edges.

    Parameters
    ----------
    beams : `~radio_beam.Beams`
        Beams object.
    tolerance : float, optional
        Allowed error range in the Khachiyan Algorithm. Decreasing the
        tolerance by an order of magnitude requires an order of magnitude
        more iterations to converge.
    nsamps : int, optional
        Number of edge points to sample from each beam.
    epsilon : float, optional
        Increase the radii of each beam by a factor of 1 + epsilon to ensure
        the common beam can marginally be deconvolved for all beams. Small
        deviations result from the finite sampling of points and the choice
        of the tolerance.
    auto_increase_epsilon : bool, optional
        Re-run the algorithm when the solution cannot quite be deconvolved from
        from all the beams. When `True`, epsilon is slightly increased with
        each iteration until the common beam can be deconvolved from all beams.
        Default is `True`.
    max_epsilon : float, optional
        Maximum epsilon value that is acceptable. Reached with `max_iter`.
        Default is `1e-3`.
    max_iter : int, optional
        Maximum number of times to increase epsilon to try finding a valid
        common beam solution.

    Returns
    -------
    com_beam : `~radio_beam.Beam`
        The common beam for all beams in the set.
    '''

    if not HAS_SCIPY:
        raise ImportError("common_manybeams_mve requires scipy.optimize.")

    step = 1

    while True:
        pts = []

        for beam in beams:
            pts.append(ellipse_edges(beam, nsamps, epsilon=epsilon))

        all_pts = np.hstack(pts).T

        # Now find the outer edges of the convex hull.
        hull = ConvexHull(all_pts)
        edge_pts = all_pts[hull.vertices]

        center, radii, rotation = \
            getMinVolEllipse(edge_pts, tolerance=tolerance)

        # The rotation matrix is coming out as:
        # ((sin theta, cos theta)
        #  (cos theta, - sin theta))
        pa = np.arctan2(- rotation[0, 0], rotation[1, 0]) * u.rad

        if pa.value == -np.pi or pa.value == np.pi:
            pa = 0.0 * u.rad

        com_beam = Beam(major=radii.max() * u.deg, minor=radii.min() * u.deg,
                        pa=pa)

        # If common beam is just slightly smaller than one of the beams,
        # we increase epsilon to encourage a solution marginally larger
        # so all beams can be convolved.
        if auto_increase_epsilon:
            if not fits_in_largest(beams, com_beam):
                # Increase epsilon and run again
                epsilon += (step + 1) * (max_epsilon - epsilon) / max_iter
                step += 1

                if step == max_iter + 1:
                    raise BeamError("Could not increase epsilon to find"
                                    " common beam.")

                continue
            else:
                break
        else:
            break

    if not fits_in_largest(beams, com_beam):
        raise BeamError("Could not find common beam to deconvolve all beams.")

    return com_beam



# this contains imports plugins that configure py.test for astropy tests.
# by importing them here in conftest.py they are discoverable by py.test
# no matter how it is invoked within the source tree.
from __future__ import print_function, absolute_import, division

import os
from distutils.version import LooseVersion

# Import casatools and casatasks here if available as they can otherwise
# cause a segfault if imported later on during tests.
try:
    import casatools
    import casatasks
except ImportError:
    pass

from astropy.version import version as astropy_version
if astropy_version < '3.0':
    # With older versions of Astropy, we actually need to import the pytest
    # plugins themselves in order to make them discoverable by pytest.
    from astropy.tests.pytest_plugins import *
else:
    # As of Astropy 3.0, the pytest plugins provided by Astropy are
    # automatically made available when Astropy is installed. This means it's
    # not necessary to import them here, but we still need to import global
    # variables that are used for configuration.
    from astropy.tests.plugins.display import PYTEST_HEADER_MODULES, TESTED_VERSIONS

from astropy.tests.helper import enable_deprecations_as_exceptions

## Uncomment the following line to treat all DeprecationWarnings as
## exceptions
# enable_deprecations_as_exceptions()

## Uncomment and customize the following lines to add/remove entries
## from the list of packages for which version numbers are displayed
## when running the tests
# try:
#     PYTEST_HEADER_MODULES['Astropy'] = 'astropy'
#     PYTEST_HEADER_MODULES['scikit-image'] = 'skimage'
#     del PYTEST_HEADER_MODULES['h5py']
# except NameError:  # needed to support Astropy < 1.0
#     pass

## Uncomment the following lines to display the version number of the
## package rather than the version number of Astropy in the top line when
## running the tests.
# import os
#
## This is to figure out the affiliated package version, rather than
## using Astropy's
# from . import version
#
# try:
#     packagename = os.path.basename(os.path.dirname(__file__))
#     TESTED_VERSIONS[packagename] = version.version
# except NameError:   # Needed to support Astropy <= 1.0.0
#     pass



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import pytest

from ..beam import Beam
from astropy.io import fits
from astropy import units as u
import os
import warnings
import numpy as np
import numpy.testing as npt
from astropy.tests.helper import assert_quantity_allclose
from itertools import product

try:
    from taskinit import ia
    HAS_CASA = True
except ImportError:
    HAS_CASA = False

from ..utils import RadioBeamDeprecationWarning, BeamError


data_dir = os.path.join(os.path.dirname(__file__), 'data')


def data_path(filename):
    return os.path.join(data_dir, filename)

def test_classic_header():
    # Instantiate from header
    fname = data_path("NGC0925.bima.mmom0.fits.gz")
    hdr = fits.getheader(fname)
    bima_beam_file = Beam.from_fits_header(fname)

    npt.assert_equal(bima_beam_file.major.value, hdr["BMAJ"])
    npt.assert_equal(bima_beam_file.minor.value, hdr["BMIN"])
    npt.assert_equal(bima_beam_file.pa.value, hdr["BPA"])

    bima_beam_hdr = Beam.from_fits_header(hdr)

    npt.assert_equal(bima_beam_hdr.major.value, hdr["BMAJ"])
    npt.assert_equal(bima_beam_hdr.minor.value, hdr["BMIN"])
    npt.assert_equal(bima_beam_hdr.pa.value, hdr["BPA"])


def test_from_aips_test():
    aips_fname = data_path("ngc0925_na.fits.gz")
    aips_hdr = fits.getheader(aips_fname)
    aips_beam_hdr = Beam.from_fits_header(aips_hdr)
    npt.assert_almost_equal(aips_beam_hdr.sr.value, 9.029858054819811e-10)
    aips_beam_file = Beam.from_fits_header(aips_fname)
    npt.assert_almost_equal(aips_beam_file.sr.value, 9.029858054819811e-10)

def test_fits_from_casa():
    casa_fname = data_path("m83.moment0.fits.gz")
    casa_hdr = fits.getheader(casa_fname)
    casa_beam_hdr = Beam.from_fits_header(casa_hdr)
    npt.assert_almost_equal(casa_beam_hdr.sr.value, 2.98323984597532e-11)
    casa_beam_file = Beam.from_fits_header(casa_fname)
    npt.assert_almost_equal(casa_beam_file.sr.value, 2.98323984597532e-11)

def test_manual():
    # Instantiate from command line
    man_beam_val = Beam(0.1, 0.1, 30)
    npt.assert_almost_equal(man_beam_val.value, 3.451589629868801e-06)
    man_beam_rad = Beam(0.1*u.rad, 0.1*u.rad, 30*u.deg)
    npt.assert_almost_equal(man_beam_rad.value, 0.011330900354567986)
    man_beam_deg = Beam(0.1*u.deg, 0.1*u.deg, 1.0*u.rad)
    npt.assert_almost_equal(man_beam_deg.value, 3.451589629868801e-06)

def test_bintable():

    beams = np.recarray(4, dtype=[('BMAJ', '>f4'), ('BMIN', '>f4'),
                                  ('BPA', '>f4'), ('CHAN', '>i4'),
                                  ('POL', '>i4')])
    beams['BMIN'] = [0.1,0.1001,0.09999,0.099999] # arcseconds
    beams['BMAJ'] = [0.2,0.2001,0.1999,0.19999]
    beams['BPA'] = [45.1,45.101,45.102,45.099] # degrees
    beams['CHAN'] = [0,0,0,0]
    beams['POL'] = [0,0,0,0]
    beams = fits.BinTableHDU(beams)

    beam = Beam.from_fits_bintable(beams)

    npt.assert_almost_equal(beam.minor.to(u.arcsec).value,
                            0.10002226, decimal=4)
    npt.assert_almost_equal(beam.major.to(u.arcsec).value,
                            0.19999751, decimal=4)
    npt.assert_almost_equal(beam.pa.to(u.deg).value,
                            45.10050065568665, decimal=4)


@pytest.mark.skipif("not HAS_CASA")
def test_from_casa_image():
    # Extract from tar
    import tarfile
    fname_tar = data_path("NGC0925.bima.mmom0.image.tar.gz")
    tar = tarfile.open(fname_tar)
    tar.extractall(path=data_dir)
    tar.close()
    fname = data_path("NGC0925.bima.mmom0.image")
    bima_casa_beam = Beam.from_casa_image(fname)


def test_attach_to_header():
    fname = data_path("NGC0925.bima.mmom0.fits.gz")
    hdr = fits.getheader(fname)
    hdr_copy = hdr.copy()
    del hdr_copy["BMAJ"], hdr_copy["BMIN"], hdr_copy["BPA"]

    bima_beam = Beam.from_fits_header(fname)

    new_hdr = bima_beam.attach_to_header(hdr_copy)

    npt.assert_equal(new_hdr["BMAJ"], hdr["BMAJ"])
    npt.assert_equal(new_hdr["BMIN"], hdr["BMIN"])
    npt.assert_equal(new_hdr["BPA"], hdr["BPA"])


def test_beam_projected_area():

    distance = 250 * u.pc

    major = 0.1 * u.rad
    beam = Beam(major, major, 30 * u.deg)

    beam_sr = (major**2 * 2 * np.pi / (8 * np.log(2))).to(u.sr)

    assert_quantity_allclose(beam_sr.value * distance ** 2,
                             beam.beam_projected_area(distance))


def test_jtok():

    major = 0.1 * u.rad
    beam = Beam(major, major, 30 * u.deg)

    freq = 1.42 * u.GHz

    conv_factor = u.brightness_temperature(beam.sr, freq)

    assert_quantity_allclose((1 * u.Jy).to(u.K, equivalencies=conv_factor),
                             beam.jtok(freq))


def test_jtok_equiv():

    major = 0.1 * u.rad
    beam = Beam(major, major, 30 * u.deg)

    freq = 1.42 * u.GHz

    conv_factor = u.brightness_temperature(beam.sr, freq)
    conv_beam_factor = beam.jtok_equiv(freq)

    assert_quantity_allclose((1 * u.Jy).to(u.K, equivalencies=conv_factor),
                             (1 * u.Jy).to(u.K, equivalencies=conv_beam_factor))

    assert_quantity_allclose((1 * u.K).to(u.Jy, equivalencies=conv_factor),
                             (1 * u.K).to(u.Jy, equivalencies=conv_beam_factor))

def test_beamarea_equiv():

    major = 0.1 * u.rad
    beam = Beam(major, major, 30 * u.deg)

    conv_factor = u.beam_angular_area(beam.sr)

    assert_quantity_allclose((1 * u.Jy / u.beam).to(u.Jy / u.sr,
                                                    equivalencies=conv_factor),
                             (1 * u.Jy / u.beam).to(u.Jy / u.sr,
                                                    equivalencies=beam.beamarea_equiv))

    assert_quantity_allclose((1 * u.Jy / u.sr).to(u.Jy / u.beam,
                                                  equivalencies=conv_factor),
                             (1 * u.Jy / u.sr).to(u.Jy / u.beam,
                                                  equivalencies=beam.beamarea_equiv))

    # Add a by-hand check
    value = (1 * u.Jy / u.sr).to(u.Jy / u.beam, equivalencies=conv_factor).value
    byhand_value = 1 * beam.sr.value
    npt.assert_allclose(value, byhand_value)

def test_convolution():

    # equations from:
    # https://github.com/pkgw/carma-miriad/blob/CVSHEAD/src/subs/gaupar.for
    # (github checkin of MIRIAD, code by Sault)

    major1 = 1 * u.deg
    minor1 = 0.5 * u.deg
    pa1 = 0.0 * u.deg
    beam1 = Beam(major1, minor1, pa1)

    major2 = 1 * u.deg
    minor2 = 0.75 * u.deg
    pa2 = 90.0 * u.deg
    beam2 = Beam(major2, minor2, pa2)

    alpha = (major1 * np.cos(pa1))**2 + (minor1 * np.sin(pa1))**2 + \
        (major2 * np.cos(pa2))**2 + (minor2 * np.sin(pa2))**2
    beta = (major1 * np.sin(pa1))**2 + (minor1 * np.cos(pa1))**2 + \
        (major2 * np.sin(pa2))**2 + (minor2 * np.cos(pa2))**2
    gamma = 2 * ((minor1**2 - major1**2) * np.sin(pa1) * np.cos(pa1) +
                 (minor2**2 - major2**2) * np.sin(pa2) * np.cos(pa2))

    s = alpha + beta
    t = np.sqrt((alpha - beta)**2 + gamma**2)

    conv_major = np.sqrt(0.5 * (s + t))
    conv_minor = np.sqrt(0.5 * (s - t))
    conv_pa = 0.5 * np.arctan2(- gamma, alpha - beta)

    conv_beam = beam1.convolve(beam2)

    assert_quantity_allclose(conv_major, conv_beam.major)
    assert_quantity_allclose(conv_minor, conv_beam.minor)
    assert_quantity_allclose(conv_pa, conv_beam.pa)


def test_deconvolution():

    # equations from:
    # https://github.com/pkgw/carma-miriad/blob/CVSHEAD/src/subs/gaupar.for
    # (github checkin of MIRIAD, code by Sault)

    major1 = 2.0 * u.deg
    minor1 = 1.0 * u.deg
    pa1 = 45.0 * u.deg
    beam1 = Beam(major1, minor1, pa1)

    major2 = 1 * u.deg
    minor2 = 0.5 * u.deg
    pa2 = 0.0 * u.deg
    beam2 = Beam(major2, minor2, pa2)

    alpha = (major1 * np.cos(pa1))**2 + (minor1 * np.sin(pa1))**2 - \
        (major2 * np.cos(pa2))**2 - (minor2 * np.sin(pa2))**2
    beta = (major1 * np.sin(pa1))**2 + (minor1 * np.cos(pa1))**2 - \
        (major2 * np.sin(pa2))**2 - (minor2 * np.cos(pa2))**2
    gamma = 2 * ((minor1**2 - major1**2) * np.sin(pa1) * np.cos(pa1) +
                 (minor2**2 - major2**2) * np.sin(pa2) * np.cos(pa2))

    s = alpha + beta
    t = np.sqrt((alpha - beta)**2 + gamma**2)

    deconv_major = np.sqrt(0.5 * (s + t))
    deconv_minor = np.sqrt(0.5 * (s - t))
    deconv_pa = 0.5 * np.arctan2(- gamma, alpha - beta)

    deconv_beam = beam1.deconvolve(beam2)

    assert_quantity_allclose(deconv_major, deconv_beam.major)
    assert_quantity_allclose(deconv_minor, deconv_beam.minor)
    assert_quantity_allclose(deconv_pa, deconv_beam.pa)


def test_conv_deconv():

    beam1 = Beam(10. * u.arcsec, 5. * u.arcsec, 30. * u.deg)
    beam2 = Beam(5. * u.arcsec, 3. * u.arcsec, 120. * u.deg)

    beam3 = beam1.convolve(beam2)

    assert beam2 == beam3.deconvolve(beam1)
    assert beam1 == beam3.deconvolve(beam2)

    assert beam1.convolve(beam2) == beam2.convolve(beam1)

    # Test multiplication and subtraction (i.e., convolution and deconvolution)
    # subtraction-as-deconvolution is deprecated. Check that one of the gives
    # the warning

    with warnings.catch_warnings(record=True) as w:
        assert beam2 == beam3 - beam1

    assert len(w) == 1
    assert w[0].category == RadioBeamDeprecationWarning
    assert str(w[0].message) == ("Subtraction-as-deconvolution is deprecated. "
                                 "Use division instead.")

    # Dividing should give the same thing
    assert beam2 == beam3 / beam1
    assert beam1 == beam3 / beam2

    assert beam3 == beam1 * beam2


@pytest.mark.parametrize(('major', 'minor', 'pa', 'return_pointlike'),
                         [[maj, min, pa, ret] for maj, min, pa, ret in
                         product([10], np.arange(1, 11),
                                 np.linspace(0, 180, 10), [True, False])])
def test_deconv_pointlike(major, minor, pa, return_pointlike):

    beam1 = Beam(major * u.arcsec, major * u.arcsec, pa * u.deg)

    if return_pointlike:
        point_beam = Beam(0 * u.deg, 0 * u.deg, 0 * u.deg)
        point_beam == beam1.deconvolve(beam1, failure_returns_pointlike=True)
    else:
        try:
            beam1.deconvolve(beam1, failure_returns_pointlike=False)
        except BeamError:
            pass


def test_isfinite():

    beam1 = Beam(10. * u.arcsec, 5. * u.arcsec, 30. * u.deg)

    assert beam1.isfinite

    # raises an exception because major < minor
    #beam2 = Beam(-10. * u.arcsec, 5. * u.arcsec, 30. * u.deg)

    #assert not beam2.isfinite

    beam3 = Beam(10. * u.arcsec, -5. * u.arcsec, 30. * u.deg)

    assert not beam3.isfinite


@pytest.mark.parametrize(("major", "minor", "pa"),
                         [(10, 10, 60),
                          (10, 10, -120),
                          (10, 10, -300),
                          (10, 10, 240),
                          (10, 10, 59),
                          (10, 10, -121)])
def test_beam_equal(major, minor, pa):

    beam1 = Beam(10 * u.deg, 10 * u.deg, 60 * u.deg)

    beam2 = Beam(major * u.deg, minor * u.deg, pa * u.deg)

    assert beam1 == beam2
    assert not beam1 != beam2


@pytest.mark.parametrize(("major", "minor", "pa"),
                         [(10, 8, 60),
                          (10, 8, -120),
                          (10, 8, 240)])
def test_beam_equal_noncirc(major, minor, pa):
    '''
    Beams with PA +/- 180 deg are equal
    '''

    beam1 = Beam(10 * u.deg, 8 * u.deg, 60 * u.deg)

    beam2 = Beam(major * u.deg, minor * u.deg, pa * u.deg)

    assert beam1 == beam2
    assert not beam1 != beam2


@pytest.mark.parametrize(("major", "minor", "pa"),
                         [(10, 8, 60),
                          (12, 10, 60),
                          (12, 10, 59)])
def test_beam_not_equal(major, minor, pa):

    beam1 = Beam(10 * u.deg, 10 * u.deg, 60 * u.deg)

    beam2 = Beam(major * u.deg, minor * u.deg, pa * u.deg)

    assert beam1 != beam2

def test_from_aips_issue43():
    """ regression test for issue 43 """
    aips_fname = data_path("header_aips.hdr")
    aips_hdr = fits.Header.fromtextfile(aips_fname)
    aips_beam_hdr = Beam.from_fits_header(aips_hdr)
    npt.assert_almost_equal(aips_beam_hdr.pa.value, -15.06)

def test_small_beam_convolution():
    # regression test for #68
    beam1 = Beam((0.1*u.arcsec).to(u.deg), (0.00001*u.arcsec).to(u.deg), 30*u.deg)
    beam2 = Beam((0.3*u.arcsec).to(u.deg), (0.00001*u.arcsec).to(u.deg), 120*u.deg)
    conv = beam1.convolve(beam2)

    np.testing.assert_almost_equal(conv.pa.to(u.deg).value, -60)

def test_major_minor_swap():

    with pytest.raises(ValueError) as exc:
        beam1 = Beam(minor=10. * u.arcsec, major=5. * u.arcsec,
                     pa=30. * u.deg)

    assert "Minor axis greater than major axis." in exc.value.args[0]



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from ._astropy_init import __version__, test

from pkg_resources import get_distribution, DistributionNotFound

from .beam import (Beam, EllipticalGaussian2DKernel,
                   EllipticalTophat2DKernel)
from .multiple_beams import Beams

__all__ = ['Beam', 'EllipticalTophat2DKernel',
           'EllipticalGaussian2DKernel', 'Beams']
